# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re


class IappParser(object):

    template_sections = [
        u'presentation',
        u'implementation',
        u'html-help',
        u'role-acl'
    ]

    template_attrs = [u'description', u'partition']

    def __init__(self, template_str):
        '''Initialize class.

        :param template_str: string of iapp template file
        :raises: EmptyTemplateException
        '''

        if template_str:
            self.template_str = unicode(template_str)
        else:
            raise EmptyTemplateException('Template empty or None value.')

    def get_section_end_index(self, section, section_start):
        '''Get end of section's content.

        :param section: string name of section
        :param section_start: integer index of section's beginning
        :return: integer index of section's end
        :raises: CurlyBraceMismatchException
        '''

        brace_count = 0

        for index, char in enumerate(self.template_str[section_start:]):
            if char == u'{':
                brace_count += 1
            elif char == u'}':
                brace_count -= 1

            if brace_count is 0:
                return index + section_start

        if brace_count is not 0:
            raise CurlyBraceMismatchException(
                'Curly braces mismatch in section %s.' % section
                )

    def get_section_start_index(self, section):
        '''Get start of a section's content.

        :param section: string name of section
        :return: integer index of section's beginning
        :raises: NonextantSectionException
        '''

        sec_start_re = '%s\s*\{' % section

        found = re.search(sec_start_re, self.template_str)
        if found:
            return found.end() - 1

        raise NonextantSectionException(
            'Section %s not found in template' % section
            )

    def get_template_name(self):
        '''Find template name.

        :returns: string of template name
        :raises: NonextantTemplateNameException
        '''

        start_pattern = 'sys application template\s+\w+\s*\{'

        template_start = re.search(start_pattern, self.template_str)
        if template_start:
            split_start = template_start.group(0).split()
            if split_start[3][-1:] == u'{':
                split_start[3] = split_start[3][:-1]
            return split_start[3]

        raise NonextantTemplateNameException('Template name not found.')

    def get_template_attr(self, attr):
        '''Find the attribute value for a specific attribute.

        :param attr: string of attribute name
        :returns: string of attribute value
        '''

        attr_re = '%s\s+.*' % attr

        attr_found = re.search(attr_re, self.template_str)
        if attr_found:
            attr_value = attr_found.group(0).replace(attr, '', 1)
            return attr_value.strip()

    def parse_template(self):
        '''Parse the template string into a dict.

        Find the (large) inner sections first, save them, and remove them from
        a modified string. Then find the template attributes in the modified
        string.

        :returns: dictionary of parsed template
        '''

        templ_dict = {}
        templ = self.template_str

        templ_dict[u'name'] = self.get_template_name()

        for section in self.template_sections:
            sec_start = self.get_section_start_index(section)
            sec_end = self.get_section_end_index(section, sec_start)
            templ_dict[section] = templ[sec_start+1:sec_end].strip()
            templ = templ[:sec_start+1] + templ[sec_end:]

        for attr in self.template_attrs:
            templ_dict[attr] = self.get_template_attr(attr)

        return templ_dict


class EmptyTemplateException(Exception):
    pass


class CurlyBraceMismatchException(Exception):
    pass


class NonextantSectionException(Exception):
    pass


class NonextantTemplateNameException(Exception):
    pass



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class VirtualCollection(Collection):
    def __init__(self, ltm):
        super(VirtualCollection, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Virtual]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:virtual:virtualstate': Virtual}


class Virtual(Resource):
    def __init__(self, virtual_collection):
        super(Virtual, self).__init__(virtual_collection)
        self._meta_data['required_json_kind'] = 'tm:ltm:virtual:virtualstate'



# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from f5.bigip.ltm.monitor import Monitor
from f5.bigip.ltm.nat import NATCollection
from f5.bigip.ltm.policy import PolicyCollection
from f5.bigip.ltm.pool import PoolCollection
from f5.bigip.ltm.rule import RuleCollection
from f5.bigip.ltm.snat import SNATCollection
from f5.bigip.ltm.virtual import VirtualCollection
from f5.bigip.resource import OrganizingCollection


class LTM(OrganizingCollection):
    def __init__(self, bigip):
        super(LTM, self).__init__(bigip)
        self._meta_data['allowed_lazy_attributes'] = [
            Monitor,
            NATCollection,
            PolicyCollection,
            PoolCollection,
            RuleCollection,
            SNATCollection,
            VirtualCollection
        ]



# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class VLANCollection(Collection):
    def __init__(self, net):
        super(VLANCollection, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [VLAN]
        self._meta_data['attribute_registry'] =\
            {'tm:net:vlan:vlanstate': VLAN}


class VLAN(Resource):
    def __init__(self, vlan_collection):
        super(VLAN, self).__init__(vlan_collection)
        self._meta_data['required_json_kind'] = 'tm:net:vlan:vlanstate'
        self._meta_data['attribute_registry'] =\
            {'tm:net:vlan:interfacescollectionstate': InterfacesCollection}


class InterfacesCollection(Collection):
    '''Represents a VLAN interface list.

    Not to be confused with tm/mgmt/net/interface.  This is actually called
    interfaces with an 's' by the BIGIP's REST API.

    *Note* We may need to come back and revist this if it turns out that
    the actual representation of the interface and interfaces objects are
    the same.
    '''
    def __init__(self, vlan):
        super(InterfacesCollection, self).__init__(vlan)
        self._meta_data['allowed_lazy_attributes'] = [Interfaces]
        self._meta_data['attribute_registry'] =\
            {'tm:net:vlan:interfaces:interfacesstate': Interfaces}


class Interfaces(Resource, ExclusiveAttributesMixin):
    def __init__(self, interfaces_collection):
        super(Interfaces, self).__init__(interfaces_collection)
        # VLAN intefaces objects do not have a partition
        self._meta_data['required_json_kind'] =\
            'tm:net:vlan:interfaces:interfacesstate'
        # You cannot send both tagged and untagged attributes on update
        self._meta_data['exclusive_attributes'].append(('tagged', 'untagged'))

    def create(self, **kwargs):
        self._create(**kwargs)
        return self

    def load(self, **kwargs):
        self._load(**kwargs)
        return self



# Copyright 2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.net.arp import ARP
from f5.bigip.net.interface import Interface
from f5.bigip.net.l2gre import L2GRE
from f5.bigip.net.route import Route
from f5.bigip.net.selfip import SelfIPCollection
from f5.bigip.net.vlan import VLANCollection
from f5.bigip.net.vxlan import VXLAN
from f5.bigip.resource import OrganizingCollection

base_uri = 'net/'


class Net(OrganizingCollection):
    def __init__(self, bigip):
        super(Net, self).__init__(bigip)
        self._meta_data['allowed_lazy_attributes'] = [
            ARP, Interface, L2GRE,
            Route, SelfIPCollection, VLANCollection, VXLAN
        ]



# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class SelfIPCollection(Collection):
    '''This represents a collection of Self IPs.

    NOTE: The objects in the collection are actually called 'self' in
    iControlREST, but obviously this will cause problems in Python so we
    changed its name to SelfIP. This is why the the SelfIPCollection's uri ends
    in self/ not selfip/.  We override the object's `_meta_data['uri']` to
    account for this.
    '''
    def __init__(self, net):
        super(SelfIPCollection, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [SelfIP]
        self._meta_data['attribute_registry'] =\
            {'tm:net:self:selfstate': SelfIP}
        # Override the URI to have self instead of the constructed selfip
        self._meta_data['uri'] =\
            self._meta_data['container']._meta_data['uri'] + "self/"


class SelfIP(Resource):
    '''This represents a Self IP.

    Use this object to create, refresh, update, delete, and load self ip
    configuration on the BIGIP.  This requires that a :class:`vlan` object
    be present on the system and that object's :attrib:`fullPath` be used
    as the vlan name.

    The address that is used for create is a *<ipaddress>/<netmask>*.  For
    example `192.168.1.1/32`.

    NOTE: The object is actully called 'self' in iControlREST, but obviously
    this will cause problems in Python so we changed its name to SelfIP.
    This is why the the SelfIPCollection's uri ends in self/ not selfip/
    '''
    def __init__(self, selfip_collection):
        super(SelfIP, self).__init__(selfip_collection)
        self._meta_data['required_json_kind'] = 'tm:net:self:selfstate'
        self._meta_data['required_creation_parameters'].update(
            ('address', 'vlan'))

    def create(self, **kwargs):
        self._create(**kwargs)



# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from f5.bigip.resource import Collection
from f5.bigip.resource import MissingRequiredCreationParameter
from f5.bigip.resource import Resource


class NATCollection(Collection):
    def __init__(self, ltm):
        super(NATCollection, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [NAT]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:nat:natstate': NAT}


class NAT(Resource):
    def __init__(self, nat_collection):
        super(NAT, self).__init__(nat_collection)
        self._meta_data['required_creation_parameters'].update(
            ('originatingAddress', 'translationAddress', 'partition'))
        self._meta_data['required_json_kind'] = 'tm:ltm:nat:natstate'

    def create(self, **kwargs):
        # If you do a create with inheritedTrafficGroup set to 'false' you
        # must also have a trafficGroup.  This pattern generalizes like so:
        # If the presence of a param implies an additional required param, then
        # simply self._meta_data['required_creation_params'].update(IMPLIED),
        # before the call to self._create(**kwargs), wherein req params are
        # checked.
        # We refer to this property as "implied-required parameters" because
        # the presence of one parameter, or parameter value (e.g.
        # inheritedTrafficGroup), implies that another parameter is required.
        itg = kwargs.get('inheritedTrafficGroup', None)
        if itg and itg == 'false':
            self._meta_data['required_creation_parameters'].\
                update(('trafficGroup',))
            try:
                if not kwargs['trafficGroup']:
                    raise MissingRequiredCreationParameter(
                        "trafficGroup must not be falsey but it's: %r"
                        % kwargs['trafficGroup'])
            except KeyError:
                pass
        self._create(**kwargs)
        return self

    def update(self, **kwargs):
        # This is an example implementation of read-only params
        stash_translation_address = self.__dict__.pop('translationAddress')
        self._update(**kwargs)
        self.__dict__['translationAddress'] = stash_translation_address



# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class FolderCollection(Collection):
    def __init__(self, sys):
        super(FolderCollection, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [Folder]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:folder:folderstate': Folder}


class Folder(Resource):
    def __init__(self, folder_collection):
        '''Create a folder resource object.

        Folder objects are the same as the partition so we need to deal with
        them slightly differently than other Resources.  For example when
        you refresh/load them you need to use the partition name instead of
        the partition and name because their self link looks something like
        `https://localhost/mgmt/tm/sys/folder/~testfolder`.  Notice that there
        is no ~partition~name format for the object.
        '''
        super(Folder, self).__init__(folder_collection)
        self._meta_data['required_json_kind'] = 'tm:sys:folder:folderstate'
        # refresh() and load() require partition, not name
        self._meta_data['required_refresh_parameters'] = set()

    def _load(self, **kwargs):
        name = kwargs.pop('name', '')
        partition = kwargs.pop('partition', '')
        read_session = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']

        if not name and not partition:
            # Root folder - https://localhost/mgmt/tm/sys/folder/~
            load_uri = base_uri + '~'
        elif not name and partition:
            # Top level - https://localhost/mgmt/tm/sys/folder/~partition
            load_uri = base_uri + '~' + partition
        elif name and not partition:
            # Top level - https://localhost/mgmt/tm/sys/folder/~partition
            load_uri = base_uri + '~' + name
        else:
            # Nested Folder (allow for name to be many folders)
            # https://localhost/mgmt/tm/sys/folder/~partition~f1~f2
            name = name.replace('/', '~')
            load_uri = base_uri + '~' + partition + '~' + name

        response = read_session.get(load_uri, uri_as_parts=False, **kwargs)
        self._local_update(response.json())
        self._build_meta_data_uri(self.selfLink)
        return self

    def update(self, **kwargs):
        '''Update the object, removing device group if inherited

        If inheritedDevicegroup is the string "true" we need to remove
        deviceGroup from the args before we update or we get the
        following error:

        The floating traffic-group: /Common/traffic-group-1 can only be set on
        /testfolder if its device-group is inherited from the root folder
        '''
        inherit_device_group = self.__dict__.get('inheritedDevicegroup', False)
        if inherit_device_group == 'true':
            self.__dict__.pop('deviceGroup')
        return self._update(**kwargs)



#!/bin/env python

# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import platform
import ssl

try:
    import StringIO
    from urllib import pathname2url
    from urllib2 import HTTPBasicAuthHandler
    from urllib2 import HTTPSHandler
    from urllib2 import ProxyHandler

except ImportError:
    from io import StringIO as StringIO  # @NoMove
    from urllib.request import HTTPBasicAuthHandler  # @UnusedImport
    from urllib.request import HTTPSHandler  # @UnusedImport
    from urllib.request import pathname2url
    from urllib.request import ProxyHandler  # @UnusedImport

from suds.cache import Cache
from suds.client import Client
from suds.client import Factory
from suds.client import ServiceSelector
from suds.options import Options
from suds.plugin import PluginContainer
from suds.reader import DefinitionsReader
from suds.servicedefinition import ServiceDefinition
from suds import transport
from suds.wsdl import Definitions
from suds.xsd.doctor import Import
from suds.xsd.doctor import ImportDoctor

# Fix missing imports. These can be global, as it applies to all f5 WSDLS.
IMP = Import('http://schemas.xmlsoap.org/soap/encoding/')
DOCTOR = ImportDoctor(IMP)
ICONTROL_URI = '/iControl/iControlPortal.cgi'
SESSION_WSDL = 'System.Session'

__version__ = '2.1'
__build__ = 'r3'


class BIGIP(object):
    """Wrap suds client object(s) and create a user-friendly class to use."""
    def __init__(self, hostname=None, username=None,
                 password=None, wsdls=None, directory=None,
                 fromurl=False, debug=False, proto='https',
                 sessions=False, cache=True, **kwargs):

        self.hostname = hostname
        self.username = username
        self.password = password
        self.directory = directory
        self.sessions = sessions
        self.fromurl = fromurl
        self.proto = proto
        self.debug = debug
        self.kw = kwargs
        self.sessionid = None

        # Setup the in-memory object cache
        if cache:
            self.cache = InMemoryCache()
        else:
            self.cache = None

        if self.debug:
            self._set_trace_logging()

        if not wsdls:
            wsdls = []

        if self.sessions:
            if SESSION_WSDL in wsdls:
                self.wsdls = [x for x in wsdls if not x.startswith(
                    'System.Session')]
                self.wsdls.insert(0, SESSION_WSDL)
            else:
                self.wsdls = wsdls
                self.wsdls.insert(0, SESSION_WSDL)
        else:
            self.wsdls = wsdls

        self.clients = self._get_clients()

        for client in self.clients:
            self._build_suds_interface(client)

    # ---------------------
    # Methods to modify active pyControl objects
    # ---------------------
    def set_timeout(self, timeout):
        if 0 < timeout <= 300:
            for client in self.clients:
                client.set_options(timeout=timeout)

    def add_interface(self, wsdl):
        if wsdl not in self.wsdls:
            self.wsdls.append(wsdl)
            client = self._get_client(wsdl)
            self._build_suds_interface(client)
            self.clients.append(client)

    def add_interfaces(self, wsdls):
        for wsdl in wsdls:
            if wsdl not in self.wsdls:
                self.wsdls.append(wsdl)
                client = self._get_client(wsdl)
                self._build_suds_interface(client)
                self.clients.append(client)

    # ---------------------
    # Setters and getters.
    # ---------------------

    def get_sessionid(self):
        """Fetch a session identifier from a v11.x BigIP."""
        mod = getattr(self, 'System')
        sessionid = getattr(mod, 'Session').get_session_identifier()
        return sessionid

    @staticmethod
    def set_sessionid(sessionid, client):
        """Sets the session header for a client.

        @sessionid (String) - session_id to add to the
                               X-iControl-Session header.
        @client - client object.

        """
        client.set_options(headers={'X-iControl-Session': sessionid})

    # ---------------------
    # Private methods
    # ---------------------

    def _build_suds_interface(self, client):
        location = '%s://%s%s' % (self.proto, self.hostname, ICONTROL_URI)

        self._set_module_attributes(client)
        self._set_interface_attributes(client)
        self._set_interface_sudsclient(client)
        self._set_type_factory(client)
        self._set_interface_methods(client)
        client.factory.separator('_')
        client.set_options(location=location, cache=self.cache)

        if self.sessions:
            if self.sessionid:
                pass
            else:
                self.sessionid = self.get_sessionid()

            self.set_sessionid(self.sessionid.__str__(), client)

    def _get_client(self, wsdl):
        url = self._set_url(wsdl)
        return self._get_suds_client(url, **self.kw)

    def _get_clients(self):
        """Get a suds client for the wsdls passed in."""
        clients = []
        for wsdl in self.wsdls:
            client = self._get_client(wsdl)
            clients.append(client)
        return clients

    @staticmethod
    def _get_module_name(c):
        """Returns the module name. Ex: 'LocalLB' """
        return c.sd[0].service.name.split('.')[0]

    def _get_module_object(self, c):
        """Returns a module object (e.g. LocalLB) """
        return getattr(self, self._get_module_name(c))

    @staticmethod
    def _get_interface_name(c):
        """Returns the interface name. Ex: 'Pool' from 'LocalLB.Pool'"""
        return c.sd[0].service.name.split('.')[1]

    def _get_interface_object(self, c, module):
        """Returns an interface object (e.g. Pool).

        @c (String) - client object
        @module - module object

        """

        return getattr(module, self._get_interface_name(c))

    @staticmethod
    def _get_methods(c):
        """Get and return a list of methods for a specific iControl interface

        """

        methods = [method[0] for method in c.sd[0].ports[0][1]]
        return methods

    def _get_suds_client(self, url, **kw):
        """Make a suds client for a specific WSDL (via url).

        Added new Suds cache features. Warning: These don't work on
        Windows. *nix should be fine. Also exposed general kwargs to
        pass down to Suds for advance users who don't want to deal
        with set_options().
        """
        if not url.startswith("https"):
            t = transport.http.HttpAuthenticated(username=self.username,
                                                 password=self.password)
            c = ROClient(url, transport=t, username=self.username,
                         password=self.password, doctor=DOCTOR, **kw)
        else:
            t = HTTPSUnVerifiedCertTransport(username=self.username,
                                             password=self.password)
            c = ROClient(url, transport=t, username=self.username,
                         password=self.password, doctor=DOCTOR, **kw)
        return c

    def _set_url(self, wsdl):
        """Set the path of file-based wsdls for processing.

        If not file-based, return a fully qualified url
        to the WSDL.
        """
        if self.fromurl:
            if wsdl.endswith('wsdl'):
                wsdl.replace('.wsdl', '')

            qstring = '?WSDL=%s' % wsdl
            return 'https://%s%s' % (self.hostname, ICONTROL_URI + qstring)
        else:
            if wsdl.endswith('wsdl'):
                pass
            else:
                wsdl += '.wsdl'

            # Check for windows and use goofy paths. Otherwise assume *nix
            if platform.system().lower() == 'windows':
                url = 'file:' + pathname2url(self.directory + '\\' + wsdl)
            else:
                url = 'file:' + pathname2url(self.directory + '/' + wsdl)
        return url

    def _set_module_attributes(self, c):
        """Sets appropriate attributes for a Module."""
        module = self._get_module_name(c)
        if hasattr(self, module):
            return
        else:
            setattr(self, module, ModuleInstance(module))

    def _set_interface_sudsclient(self, c):
        """Set an attribute that points to the actual suds client.

        This will allow for power-users to get at suds client internals.
        """
        module = self._get_module_object(c)
        interface = self._get_interface_object(c, module)
        setattr(interface, 'suds', c)

    def _set_interface_attributes(self, c):
        """Sets appropriate attributes for a Module."""
        module = self._get_module_object(c)
        interface = self._get_interface_name(c)
        setattr(module, interface, InterfaceInstance(interface))

    def _set_interface_methods(self, c):
        """Sets up methods as attributes for a particular iControl interface.

        Method keys (attrs) point to suds.service objects for the interface.
        """
        module = self._get_module_object(c)
        interface = self._get_interface_object(c, module)
        methods = self._get_methods(c)

        for method in methods:
            suds_method = getattr(c.service, method)
            setattr(interface, method, suds_method)
            m = getattr(interface, method)
            self._set_method_input_params(c, m, method)
            self._set_return_type(c, m, method)

    @staticmethod
    def _set_method_input_params(c, interface_method, method):
        """Set the method input argument attribute named 'params' for easy reference.

        """

        m = c.sd[0].ports[0][0].method(method)
        params = []
        for x in m.soap.input.body.parts:
            params.append((x.name, x.type[0]))
        setattr(interface_method, 'params', params)

    @staticmethod
    def _set_return_type(c, interface_method, method):
        """Sets the return type in an attribute named response_type"""
        m = c.sd[0].ports[0][0].method(method)
        if len(m.soap.output.body.parts):
            res = m.soap.output.body.parts[0].type[0]
            setattr(interface_method, 'response_type', res)
        else:
            setattr(interface_method, 'response_type', None)

    def _set_type_factory(self, c):
        factory = getattr(c, 'factory')
        module = self._get_module_object(c)
        interface = self._get_interface_object(c, module)
        setattr(interface, 'typefactory', factory)

    @staticmethod
    def _set_trace_logging():
        logging.basicConfig(level=logging.INFO)
        logging.getLogger('suds.client').setLevel(logging.DEBUG)


class ModuleInstance(object):
    """An iControl module object to set attributes against."""
    def __init__(self, name):
        self.name = name


class InterfaceInstance(object):
    """An iControl interface object to set attributes against."""
    def __init__(self, name):
        self.name = name


class ROClient(Client):
    def __init__(self, url, **kwargs):
        """ROClient

        @param url: The URL for the WSDL.
        @type url: str
        @param kwargs: keyword arguments.
        @see: L{Options}
        """
        options = Options()
        options.transport = transport.https.HttpAuthenticated()
        self.options = options
        options.cache = InMemoryCache()
        self.set_options(**kwargs)
        reader = DefinitionsReader(options, Definitions)
        self.wsdl = reader.open(url)
        plugins = PluginContainer(options.plugins)
        plugins.init.initialized(wsdl=self.wsdl)
        self.factory = Factory(self.wsdl)
        self.service = ServiceSelector(self, self.wsdl.services)
        self.sd = []
        for s in self.wsdl.services:
            sd = ServiceDefinition(self.wsdl, s)
            self.sd.append(sd)
        self.messages = dict(tx=None, rx=None)


class InMemoryCache(Cache):
    """In-memory cache.

    The contents of the cache is shared between all instances.
    """
    data = {}

    def get(self, objid):
        return self.data.get(objid)

    def getf(self, objid):
        obj = self.get(objid)
        return StringIO(obj) if obj else None

    def put(self, objid, obj):
        self.data[objid] = obj

    def putf(self, objid, fp):
        self.put(fp.read())

    def purge(self, objid):
        del self.data[objid]

    def clear(self):
        self.data = {}


class HTTPSUnVerifiedCertTransport(transport.https.HttpAuthenticated):

    def __init__(self, *args, **kwargs):
        transport.https.HttpAuthenticated.__init__(self, *args, **kwargs)

    def u2handlers(self):
        handlers = []
        handlers.append(ProxyHandler(self.proxy))
        handlers.append(HTTPBasicAuthHandler(self.pm))
        # python ssl Context support - PEP 0466
        if hasattr(ssl, '_create_unverified_context'):
            ssl_context = ssl._create_unverified_context()
            handlers.append(HTTPSHandler(context=ssl_context))
        else:
            handlers.append(HTTPSHandler())
        return handlers


def main():
    import sys
    if len(sys.argv) < 4:
        print("Usage: %s <hostname> <username> <password>" % sys.argv[0])
        sys.exit()

    a = sys.argv[1:]
    b = BIGIP(hostname=a[0],
              username=a[1],
              password=a[2],
              fromurl=True,
              wsdls=['LocalLB.Pool'])

    pools = b.LocalLB.Pool.get_list()
    version = b.LocalLB.Pool.get_version()
    print("Version is: %s\n" % version)
    print("Pools:")
    for x in pools:
        print("\t%s" % x)

if __name__ == '__main__':
    main()



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# CONSTANTS MODULE
MIN_TMOS_MAJOR_VERSION = 11
MIN_TMOS_MINOR_VERSION = 5
MIN_EXTRA_MB = 500
DEFAULT_HOSTNAME = 'bigip1'
MAX_HOSTNAME_LENGTH = 128
DEFAULT_FOLDER = "Common"
FOLDER_CACHE_TIMEOUT = 120
CONNECTION_TIMEOUT = 30
FDB_POPULATE_STATIC_ARP = True
# DEVICE LOCK PREFIX
DEVICE_LOCK_PREFIX = 'lock_'
# DIR TO CACHE WSDLS.  SET TO NONE TO READ FROM DEVICE
# WSDL_CACHE_DIR = "/data/iControl-11.4.0/sdk/wsdl/"
WSDL_CACHE_DIR = ''
# HA CONSTANTS
HA_VLAN_NAME = "HA"
HA_SELFIP_NAME = "HA"
# VIRTUAL SERVER CONSTANTS
VS_PREFIX = 'vs'
# POOL CONSTANTS
POOL_PREFIX = 'pool'
# POOL CONSTANTS
MONITOR_PREFIX = 'monitor'
# VLAN CONSTANTS
VLAN_PREFIX = 'vlan'
# BIG-IP PLATFORM CONSTANTS
BIGIP_VE_PLATFORM_ID = 'Z100'
# DEVICE CONSTANTS
DEVICE_DEFAULT_DOMAIN = ".local"
DEVICE_HEALTH_SCORE_CPU_WEIGHT = 1
DEVICE_HEALTH_SCORE_MEM_WEIGHT = 1
DEVICE_HEALTH_SCORE_CPS_WEIGHT = 1
DEVICE_HEALTH_SCORE_CPS_PERIOD = 5
DEVICE_HEALTH_SCORE_CPS_MAX = 100
# DEVICE GROUP CONSTANTS
PEER_ADD_ATTEMPTS_MAX = 10
PEER_ADD_ATTEMPT_DELAY = 2
DEFAULT_SYNC_MODE = 'autosync'
# MAX RAM SYNC DELAY IS 63 SECONDS
# (3+6+9+12+15+18) = 63
SYNC_DELAY = 3
MAX_SYNC_ATTEMPTS = 10
# SHARED CONFIG CONSTANTS
SHARED_CONFIG_DEFAULT_TRAFFIC_GROUP = 'traffic-group-local-only'
SHARED_CONFIG_DEFAULT_FLOATING_TRAFFIC_GROUP = 'traffic-group-1'
VXLAN_UDP_PORT = 4789



class LazyAttributeMixin(object):
    def __getattr__(self, name):
        for lazy_attribute in self.allowed_lazy_attributes:
            if name == lazy_attribute.__name__.lower():
                iface_collection = lazy_attribute(self)
                setattr(self, name, iface_collection)
                return iface_collection
        error_message = "'%s' object has no attribute '%s'"\
            % (self.__class__, name)
        raise AttributeError(error_message)



# Copyright 2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import OrganizingCollection
from f5.bigip.sys.application import ApplicationCollection
from f5.bigip.sys.folder import FolderCollection
from f5.bigip.sys.stat import Stat
from f5.bigip.sys.system import System


base_uri = 'sys/'


class Sys(OrganizingCollection):
    def __init__(self, bigip):
        super(Sys, self).__init__(bigip)
        self._meta_data['allowed_lazy_attributes'] = [
            FolderCollection,
            ApplicationCollection, Stat, System
        ]



"""Classes and functions for configuring BIG-IP"""
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os

from f5.bigip.cm import CM
from f5.bigip.cm.device import Device
from f5.bigip.ltm import LTM
from f5.bigip.net import Net
from f5.bigip.pycontrol import pycontrol as pc
from f5.bigip.resource import OrganizingCollection
from f5.bigip.sys import Sys
from f5.common import constants as const
from icontrol.session import iControlRESTSession

LOG = logging.getLogger(__name__)
allowed_lazy_attributes = [CM, Device, LTM, Net, Sys]


def _get_icontrol(hostname, username, password, timeout=None):
    """Initialize iControl interface"""
    # Logger.log(Logger.DEBUG,
    #           "Opening iControl connections to %s for interfaces %s"
    #            % (self.hostname, self.root_collections))

    if os.path.exists(const.WSDL_CACHE_DIR):
        icontrol = pc.BIGIP(hostname=hostname,
                            username=username,
                            password=password,
                            directory=const.WSDL_CACHE_DIR,
                            wsdls=[])
    else:
        icontrol = pc.BIGIP(hostname=hostname,
                            username=username,
                            password=password,
                            fromurl=True,
                            wsdls=[])

    if timeout:
        icontrol.set_timeout(timeout)
    else:
        icontrol.set_timeout(const.CONNECTION_TIMEOUT)

    return icontrol


class BigIP(OrganizingCollection):
    """An interface to a single BIG-IP"""
    def __init__(self, hostname, username, password, **kwargs):
        timeout = kwargs.pop('timeout', 30)
        allowed_lazy_attrs = kwargs.pop('allowed_lazy_attributes',
                                        allowed_lazy_attributes)
        if kwargs:
            raise TypeError('Unexpected **kwargs: %r' % kwargs)
        # _meta_data variable values
        iCRS = iControlRESTSession(username, password, timeout=timeout)
        icontrol_inst = _get_icontrol(hostname, username, password)
        # define _meta_data
        self._meta_data = {'allowed_lazy_attributes': allowed_lazy_attrs,
                           'icontrol': icontrol_inst,
                           'hostname': hostname,
                           'uri': 'https://%s/mgmt/tm/' % hostname,
                           'icr_session': iCRS,
                           'device_name': None,
                           'local_ip': None,
                           'bigip': self}



"""BIG-IP API RESTInterfaceCollection """
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os

from f5.bigip import exceptions
from f5.common import constants as const
from f5.common.logger import Log
from requests.exceptions import HTTPError

OBJ_PREFIX = 'uuid_'

LOG = logging.getLogger(__name__)


def log(method):
    """Decorator helping to log method calls."""
    def wrapper(*args, **kwargs):
        """Necessary wrapper """
        instance = args[0]
        LOG.debug('%s::%s called with args: %s kwargs: %s',
                  instance.__class__.__name__,
                  method.__name__,
                  args[1:],
                  kwargs)
        return method(*args, **kwargs)
    return wrapper


class RESTInterfaceCollection(object):
    """Abstract base class for collection objects. """
    @log
    def delete(self, name=None, folder='Common',
               timeout=const.CONNECTION_TIMEOUT):
        """Delete the object """
        if not name:
            return False
        try:
            self.bigip.icr_session.delete(
                self.base_uri,
                folder=folder,
                instance_name=name,
                timeout=timeout)
        except HTTPError as exp:
            if exp.response.status_code == 404:
                return True
            else:
                raise
        return True

    def _delete(self, folder, name, timeout):
        try:
            self.bigip.icr_session.delete(self.base_uri, folder, name, timeout)
        except HTTPError as err:
            if (err.response.status_code == 400
                    and err.response.text.find('is referenced') > 0):
                Log.error('members', err.response.text)
            else:
                raise exceptions.PoolDeleteException(err.response.text)
        else:
            self._del_arp_and_fdb(name, folder)

    @log
    def delete_all(self, folder='Common', startswith="",
                   timeout=const.CONNECTION_TIMEOUT):
        """Delete all things that can start with a string.

        Used to use self.OBJ_PREFIX so now you have to pass it in.
        Maybe this isn't the best thing?

        We need to deal with the prefix in a better way
        """
        params = {
            '$select': 'name,selfLink',
            '$filter': 'partition eq ' + folder,
        }

        # This will raise if there is a HTTPError
        response = self.bigip.icr_session.get(
            self.base_uri, params=params, timeout=timeout)

        items = response.json().get('items', [])
        for item in items:
            # This is where we had startswith(self.OBJ_PREFIX)
            if item['name'].startswith(startswith):
                if not self.delete(item['name'], folder=folder):
                    return False
        return True

    @log
    def exists(self, name=None, folder='Common',
               timeout=const.CONNECTION_TIMEOUT):
        try:
            self.bigip.icr_session.get(
                self.base_uri, folder=folder, instance_name=name,
                params={'$select': 'name'}, timeout=timeout)
        except HTTPError as exp:
            if exp.response.status_code == 404:
                return False
            else:
                raise
        return True

    @log
    def _get_items(self, folder='Common', name='', suffix='/members',
                   select='name', timeout=const.CONNECTION_TIMEOUT, **kwargs):
        items = []
        params = {
            '$select': select,
            '$filter': 'partition eq ' + folder
        }
        try:
            response = self.bigip.icr_session.get(
                self.base_uri, folder, name, params=params,
                timeout=timeout, **kwargs)
        except HTTPError as exp:
            if exp.response.status_code == 404:
                return items
            raise

        items = response.json().get('items', [])
        if select:
            for item in items:
                if select in item:
                    items.append(strip_folder_and_prefix(item[select]))

        return items

    @log
    def _get_named_object(self, name, uri=None, folder='Common', select='name',
                          timeout=const.CONNECTION_TIMEOUT):
        params = {
            '$select': select,
        }
        if not uri:
            uri = self.base_uri

        # No try here because original code was not doing exceptional things
        # with error messages like self._get()
        response = self.bigip.icr_session.get(uri,
                                              instance_name=name,
                                              folder=folder,
                                              params=params,
                                              timeout=timeout)
        return response.json().get(select, None)

    @log
    def _set(self, name, folder, key, value):
        payload = {key: value}
        try:
            self.bigip.icr_session.put(
                self.base_uri,
                instance_name=name,
                folder=folder,
                json=payload,
                timeout=const.CONNECTION_TIMEOUT)
        except HTTPError as err:
            Log.error(self.__class__.__name__, err.response.text)
            raise
        return True


def prefixed(name):
    """Put object prefix in front of name """
    if not name.startswith(OBJ_PREFIX):
        name = OBJ_PREFIX + name
    return name


def icontrol_folder(method):
    """Returns the iControl folder + object name.

    If a kwarg name is 'name' or else ends in '_name'.

    The folder and the name will be prefixed with the global
    prefix OBJ_PREFIX. If preserve_vlan_name=True is an argument,
    then the 'vlan_name' argument will not be prefixed but the
    other matching arguments will.

    It also sets the iControl active folder to folder kwarg
    assuring get_list returns just the appopriate objects
    for the specific administrative partition. It does this
    for kwarg named 'name', ends in '_name', or 'named_address'.

    If the value in the name already includes '/Common/' the
    decoration honors that full path.
    """
    def wrapper(*args, **kwargs):
        """Necessary wrapper """
        instance = args[0]
        preserve_vlan_name = False
        if 'preserve_vlan_name' in kwargs:
            preserve_vlan_name = kwargs['preserve_vlan_name']
        if 'folder' in kwargs and kwargs['folder']:
            if kwargs['folder'].find('~') > -1:
                kwargs['folder'] = kwargs['folder'].replace('~', '/')
            kwargs['folder'] = os.path.basename(kwargs['folder'])
            if not kwargs['folder'] == 'Common':
                kwargs['folder'] = prefixed(kwargs['folder'])
            if 'name' in kwargs and kwargs['name']:
                if isinstance(kwargs['name'], basestring):
                    if kwargs['name'].find('~') > -1:
                        kwargs['name'] = kwargs['name'].replace('~', '/')
                    if kwargs['name'].startswith('/Common/'):
                        kwargs['name'] = os.path.basename(kwargs['name'])
                        kwargs['name'] = prefixed(kwargs['name'])
                        kwargs['name'] = instance.bigip.set_folder(
                            kwargs['name'], 'Common')
                    else:
                        kwargs['name'] = os.path.basename(kwargs['name'])
                        kwargs['name'] = prefixed(kwargs['name'])
                        kwargs['name'] = instance.bigip.set_folder(
                            kwargs['name'], kwargs['folder'])
            if 'named_address' in kwargs and kwargs['named_address']:
                if isinstance(kwargs['name'], basestring):
                    if kwargs['named_address'].find('~') > -1:
                        kwargs['named_address'] = \
                            kwargs['named_address'].replace('~', '/')
                    if kwargs['named_address'].startswith('/Common/'):
                        kwargs['named_address'] = \
                            os.path.basename(kwargs['named_address'])
                        kwargs['named_address'] = \
                            instance.bigip.set_folder(kwargs['named_address'],
                                                      'Common')
                    else:
                        kwargs['named_address'] = \
                            os.path.basename(kwargs['named_address'])
                        kwargs['named_address'] = \
                            instance.bigip.set_folder(kwargs['named_address'],
                                                      kwargs['folder'])
            for name in kwargs:
                if name.find('_folder') > 0 and kwargs[name]:
                    if kwargs[name].find('~') > -1:
                        kwargs[name] = kwargs[name].replace('~', '/')
                    kwargs[name] = os.path.basename(kwargs[name])
                    if not kwargs[name] == 'Common':
                        kwargs[name] = prefixed(kwargs[name])
                if name.find('_name') > 0 and kwargs[name]:
                    if isinstance(kwargs['name'], basestring):
                        if kwargs[name].find('~') > -1:
                            kwargs[name] = kwargs[name].replace('~', '/')
                        if kwargs[name].startswith('/Common/'):
                            kwargs[name] = os.path.basename(kwargs[name])
                            if name != 'vlan_name' or not preserve_vlan_name:
                                kwargs[name] = prefixed(kwargs[name])
                            kwargs[name] = instance.bigip.set_folder(
                                kwargs[name], 'Common')
                        else:
                            name_prefix = name[0:name.index('_name')]
                            specific_folder_name = name_prefix + "_folder"
                            folder = kwargs['folder']
                            if specific_folder_name in kwargs:
                                folder = kwargs[specific_folder_name]
                            kwargs[name] = os.path.basename(kwargs[name])
                            if name != 'vlan_name' or not preserve_vlan_name:
                                kwargs[name] = prefixed(kwargs[name])
                            kwargs[name] = instance.bigip.set_folder(
                                kwargs[name], folder)
            instance.bigip.set_folder(None, kwargs['folder'])
        return method(*args, **kwargs)
    return wrapper


def icontrol_rest_folder(method):
    """Returns iControl REST folder + object name.

    If a kwarg name is 'name' or else ends in '_name'.

    The folder and the name will be prefixed with the global
    prefix OBJ_PREFIX.
    """
    def wrapper(*args, **kwargs):
        """Necessary wrapper """
        preserve_vlan_name = False
        if 'preserve_vlan_name' in kwargs:
            preserve_vlan_name = kwargs['preserve_vlan_name']

        # Here we make sure the name or folder is not REST formatted,
        # which uses '~' instead of '/'. We change them back to '/'.
        # We normalize the object names to their base name (with no
        # / in the name at all) and then use a common prefix.
        if 'folder' in kwargs and kwargs['folder']:
            if kwargs['folder'] != '/' and kwargs['folder'].find('Common') < 0:
                temp = kwargs['folder'].replace('~', '/')
                kwargs['folder'] = prefixed(os.path.basename(temp))
        if 'name' in kwargs and kwargs['name']:
            if isinstance(kwargs['name'], basestring):
                temp = kwargs['name'].replace('~', '/')
                kwargs['name'] = prefixed(os.path.basename(temp))
            else:
                LOG.warn('attempting to normalize non basestring name. '
                         'Argument: val: ' + str(kwargs['name']))

        for name in kwargs:
            if name.find('_folder') > 0 and kwargs[name]:
                kwargs[name] = kwargs[name].replace('~', '/')
                kwargs[name] = os.path.basename(kwargs[name])
                if not kwargs[name] == 'Common':
                    kwargs[name] = prefixed(kwargs[name])
            if name.find('_name') > 0 and kwargs[name]:
                if isinstance(kwargs[name], basestring):
                    kwargs[name] = kwargs[name].replace('~', '/')
                    kwargs[name] = os.path.basename(kwargs[name])
                    if name != 'vlan_name' or not preserve_vlan_name:
                        kwargs[name] = prefixed(kwargs[name])
                else:
                    LOG.warn('attempting to normalize non basestring name. '
                             ' Argument: name: ' + str(name) +
                             ' val:' + str(kwargs[name]))
        return method(*args, **kwargs)
    return wrapper


def decorate_name(name=None, folder='Common', use_prefix=True):
    """Add "namespace" prefix to names """
    folder = os.path.basename(folder)
    if not folder == 'Common':
        folder = prefixed(folder)
    if name.startswith('/Common/'):
        name = os.path.basename(name)
        if use_prefix:
            name = prefixed(name)
        name = '/Common/' + name
    else:
        name = os.path.basename(name)
        if use_prefix:
            name = prefixed(name)
        name = '/' + folder + '/' + name
    return name


def strip_folder_and_prefix(path):
    """Strip folder and prefix """
    if isinstance(path, list):
        for i in range(len(path)):
            if path[i].find('~') > -1:
                path[i] = path[i].replace('~', '/')
            if path[i].startswith('/Common'):
                path[i] = path[i].replace(OBJ_PREFIX, '')
            else:
                path[i] = \
                    os.path.basename(str(path[i])).replace(OBJ_PREFIX, '')
        return path
    else:
        if path.find('~') > -1:
            path = path.replace('~', '/')
        if path.startswith('/Common'):
            return str(path).replace(OBJ_PREFIX, '')
        else:
            return os.path.basename(str(path)).replace(OBJ_PREFIX, '')


def strip_domain_address(ip_address):
    """Strip domain from ip address """
    mask_index = ip_address.find('/')
    if mask_index > 0:
        return ip_address[:mask_index].split('%')[0] + ip_address[mask_index:]
    else:
        return ip_address.split('%')[0]


def split_addr_port(dest):
    if len(dest.split(':')) > 2:
        # ipv6: bigip syntax is addr.port
        parts = dest.split('.')
    else:
        # ipv4: bigip syntax is addr:port
        parts = dest.split(':')
    return (parts[0], parts[1])



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import suds
from xml.sax import SAXParseException

# Project info


class F5Error(Exception):
    def __init__(self, e):
        self.exception = e
        self.msg = str(e)

        if isinstance(e, suds.WebFault):
            try:
                parts = e.fault.faultstring.split('\n')
                # e_source = parts[0].replace("Exception caught in ", "")
                e_type = parts[1].replace("Exception: ", "")
                e_msg = re.sub("\serror_string\s*:\s*", "", parts[4])
                self.msg = "%s: %s" % (e_type, e_msg)
            except IndexError:
                self.msg = e.fault.faultstring
        if isinstance(e, SAXParseException):
            self.msg = "Unexpected server response. %s" % e.message

    def __str__(self):
        return self.msg



# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""This module provides classes that specify how RESTful resources are handled.

There are different types of resources published by the BigIP REST Server, they
are represented by the classes in this module.

Available Classes:
    * ResourceBase -- only `read` is generally supported in all resource types,
      this class provides `read`. ResourceBase objects are usually instantiated
      via setting lazy attributes. ResourceBase provides a contructor to match
      the lazy constructor. The expected behavior is that all resource
      subclasses depend on this constructor to correctly set their
      self._meta_data['uri'].
      All ResourceBase objects (except BigIPs) have a container (BigIPs contain
      themselves).  The container is the object the ResourceBase is an
      attribute of.
    * Collection -- These resources support lists of ResourceBase Objects.
    * Resource -- These resources are the only resources that support
      `create`, `update`, and `delete` operations.  Because they support HTTP
      post (via _create) they uniquely depend on 2 uri's, a uri that supports
      the creating post, and the returned uri of the newly created resource.
    * InvalidResource -- resources do not generally support all 5 Resource
      operations, if a caller attempts to invoke an unsupported operation this
      Exception is raised.
"""
import functools
import urlparse

from f5.bigip.mixins import LazyAttributeMixin
from f5.bigip.mixins import ToDictMixin


class KindTypeMismatch(Exception):
    pass


class DeviceProvidesIncompatibleKey(Exception):
    pass


class InvalidResource(Exception):
    """Raise this when a caller tries to invoke an unsupported CUD op.

    All resources support `read`.
    Only Resources support `create`, `update`, and `delete`.
    """
    pass


class MissingRequiredCreationParameter(Exception):
    pass


class MissingRequiredReadParameter(Exception):
    pass


class UnregisteredKind(Exception):
    pass


class GenerationMismatch(Exception):
    pass


class InvalidForceType(ValueError):
    pass


class URICreationCollision(Exception):
    pass


class ResourceBase(LazyAttributeMixin, ToDictMixin):
    """Every resource that maps to a uri on the device should inherit this.

    Instantiate this with ContainerInstance.NewResourceBaseInstance via the
    LazyAttributeMixin.

    The BigIP is represented by an object that converts device published uri's
    into Python objects.  Each uri maps to a Python object. The mechanism for
    instantatiating these objects is the __getattr__ Special Function in the
    LazyAttributeMixin.  When a registered attribute is `dot` referenced, on
    the device object (e.g. `bigip.ltm` or simply `bigip`), an appropriate
    object is instantiated and attributed to the referencing object, so:
    >>> bigip.ltm = LTM(bigip)
    >>> bigip.ltm.natcollection
    >>> nat1 = bigip.ltm.natcollection.nat.create('Foo', 'Bar', '0.1.2.3',
                                                  '1.2.3.4')

    Shortens to just the last line:
    >>> nat1 = bigip.ltm.natcollection.nat.create('Foo', 'Bar', '0.1.2.3',
                                                  '1.2.3.4')

    More importantly is enforces a convention relating device published uris to
    API objects, in a hierarchy similar to the uri paths.  I.E. the uri
    corresponding to a `NatCollection` object is `mgmt/tm/ltm/nat/`. If you
    query the bigip's uri (e.g. print(bigip._meta_data['uri']) ), you'll see
    that it ends in:
    `mgmt/tm/`, if you query the `ltm` object's uri
    (e.g. print(bigip.ltm._meta_data['uri']) ) you'll see it ends in
    `mgmt/tm/ltm/.

    In general the objects build a required `self._meta_data['uri']` attribute
    by:
    1. Inheriting this class.
    2. calling super(Subclass, self).__init__(container)
    3. self.uri = self.container_uri + base_uri <-- Always defined in the
    module that defines the class.

    The net result is a reasonably succinct mapping between uri's and objects,
    that represents objects in a hierarchical relationship similar to the
    devices uri path hierarchy.
    This is the class that defines `read` for all ResourceBases.

    Public attributes:
    - read: updates itself with the results of an http GET on the resource
    - others: not to be called here

    """
    def __init__(self, container):
        """Call this with containing_object_instance.FOO

        FOO must inherit from this class.  The '.' operator passes "FOO" to
        the __getattr__ method of the containing_object_instance where it is
        instantiated as the appropriate type of ResourceBase.

        Since all ResourceBases support `read` instances of ResourceBase do as
        well.
        The BigIP uri 'mgmt/tm/' uniquely passes itself to this constructor as
        the "container".
        """
        self._meta_data = {'container': container,
                           'bigip': container._meta_data['bigip']}

    def _local_update(self, rdict):
        sanitized = self._check_keys(rdict)
        temp_meta = self._meta_data
        self.__dict__ = sanitized
        self._meta_data = temp_meta

    def _check_keys(self, rdict):
        if '_meta_data' in rdict:
            error_message = "Response contains key '_meta_data' which is " +\
                "incompatible with this API!!\n Response json: %r" % rdict
            raise DeviceProvidesIncompatibleKey(error_message)
        for x in rdict:
            if x.startswith('__'):
                raise DeviceProvidesIncompatibleKey(x)
        return rdict

    def _refresh(self):
        """Use this to make the device resource be represented by self.

        This method is run for its side-effects on self.
        This method makes an HTTP get query against its uri, if successful its
        attribute __dict__ is replaced with the dict representing the device
        state.  To figure out what that state is, run a subsequest query of the
        object like this:
        >>> resource_obj.read()
        >>> print(resource.name)
        """
        read_session = self._meta_data['bigip']._meta_data['icr_session']
        response = read_session.get(self._meta_data['uri'])
        self._local_update(response.json())

    def refresh(self):
        self._refresh()

    def _build_meta_data_uri(self, selfLinkuri):
        hostname = self._meta_data['bigip']._meta_data['hostname']
        (scheme, domain, path, qarg, frag) = urlparse.urlsplit(selfLinkuri)
        path_uri = urlparse.urlunsplit((scheme, hostname, path, '', ''))+'/'
        self._meta_data['uri'] = path_uri
        self._meta_data['creation_uri_qarg'] = qarg
        self._meta_data['creation_uri_frag'] = frag

    def load(self, **kwargs):
        error_message = "Only Resources support 'load'."
        raise InvalidResource(error_message)

    def create(self):
        error_message = "Only Resources support 'create'."
        raise InvalidResource(error_message)

    def update(self):
        error_message = "Only Resources support 'update'."
        raise InvalidResource(error_message)

    def delete(self):
        error_message = "Only Resources support 'delete'."
        raise InvalidResource(error_message)

    @property
    def raw(self):
        return self.__dict__


class OrganizingCollection(ResourceBase):
    def __init__(self, bigip):
        super(OrganizingCollection, self).__init__(bigip)
        base_uri = self.__class__.__name__.lower() + '/'
        self._meta_data['uri'] =\
            self._meta_data['container']._meta_data['uri'] + base_uri
        # Collections have a registry which must be reified in the
        # subclass constructor.
        self._meta_data['attribute_registry'] = {}

    # Because of the behavior of the BigIP REST server different resource types
    # must handle get_collection differently.
    def get_collection(self):
        self._refresh()
        return self.items


class Collection(ResourceBase):
    """Inherit from this class if the corresponding uri lists other resources.

    Note any subclass must have "Collection" at the end of its name!
    """
    def __init__(self, container):
        super(Collection, self).__init__(container)
        # Handle 'collection'
        endind = len('collection')
        base_uri = self.__class__.__name__.lower()[:-endind] + '/'
        self._meta_data['uri'] =\
            self._meta_data['container']._meta_data['uri'] + base_uri

    def get_collection(self):
        """Get an iterator (list maybe upgrade to generator) of objects.

        The objects in returned list are Pythonic ResourceBases that map to the
        most recently `got` state of uris-resources published by the device.
        In order to instantiate the correct types, the concrete subclass must
        populate its registry with acceptable types, based on the `kind` field
        returned by the REST server.
        """
        list_of_contents = []
        # Collections list is likely to become collections.abc.Sequence subtype
        # with support for field based comparison.
        self._refresh()
        if 'items' in self.__dict__:
            for item in self.items:
                if 'kind' not in item:
                    list_of_contents.append(item)
                    continue
                kind = item['kind']
                if kind in self._meta_data['attribute_registry']:
                    instance =\
                        self._meta_data['attribute_registry'][kind](self)
                    instance._local_update(item)
                    instance._build_meta_data_uri(instance.selfLink)
                    list_of_contents.append(instance)
                else:
                    error_message = '%r is not registered!' % kind
                    raise UnregisteredKind(error_message)
        return list_of_contents


class Resource(ResourceBase):
    """Use this to represent a Configurable Resource on the device.

    1a.  bigip.ltm.natcollection.nat
    or
    1b.  nat_obj = bigip.ltm.natcollection.nat
    2.  call super(Subclass, self).__init__(container) in its __init__
    """
    def __init__(self, container):
        """XXX

        Call _create for a Resource resource to have a self._meta_data['uri']!
        """
        super(Resource, self).__init__(container)
        # All Creation supporting Resources must update the
        # 'required_creation_parameters' set with the appropriate values.
        self._meta_data['required_creation_parameters'] = set(('name',))
        self._meta_data['required_refresh_parameters'] = set(('name',))
        self._meta_data['exclusive_attributes'] = []
        self._meta_data['read_only_attributes'] = []

    def _manage_local_creation(decorated):
        @functools.wraps(decorated)
        def wrapper(instance, **kwargs):
            if 'uri' in instance._meta_data:
                error = "There was an attempt to assign a new uri to this "\
                        "resource, the _meta_data['uri'] is %s and it should"\
                        " not be changed." % (instance._meta_data['uri'])
                raise URICreationCollision(error)
            returned = decorated(instance, **kwargs)
            attribute_reg = instance._meta_data.get('attribute_registry', {})
            attrs = attribute_reg.values()
            instance._meta_data['allowed_lazy_attributes'] = attrs
            return returned
        return wrapper

    @_manage_local_creation
    def _create(self, **kwargs):
        """Call this to create.

        Subclasses, should support this functionality by defining a `create`
        method that wraps and calls this method with appropriate arguments.

        :params kwargs: All the key-values needed to create the resource
        :returns: An instance of the Python object that represents the device's
        uri-published resource.  The uri of the resource is part of the
        object's _meta_data.
        :returns: Note this is the only fundamental Resource operation that
        returns a different uri (in the returned object) than the uri the
        operation was called on.  The returned uri can be accessed as
        Object.selfLink, the actual uri used by REST operations on the object
        is Object._meta_data['uri'].  The _meta_data['uri'] is the same as
        Object.selfLink with the substring 'localhost' replaced with the value
        of Object._meta_data['bigip']._meta_data['hostname'].
        """

        key_set = set(kwargs.keys())
        required_minus_received =\
            self._meta_data['required_creation_parameters'] - key_set
        if required_minus_received != set():
            error_message = 'Missing required params: %r'\
                % required_minus_received
            raise MissingRequiredCreationParameter(error_message)

        # Make convenience variable with short names for this method.
        _create_uri = self._meta_data['container']._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']

        # Invoke the REST operation on the device.
        response = session.post(_create_uri, json=kwargs)

        # Post-process the response
        self._local_update(response.json())

        if self.kind != self._meta_data['required_json_kind']:
            error_message = "For instances of type '%r' the corresponding"\
                " kind must be '%r' but creation returned JSON with kind: %r"\
                % (self.__class__.__name__,
                   self._meta_data['required_json_kind'],
                   self.kind)
            raise KindTypeMismatch(error_message)

        # Update the object to have the correct functional uri.
        self._build_meta_data_uri(self.selfLink)
        return self

    def create(self, **kwargs):
        self._create(**kwargs)
        return self

    @_manage_local_creation
    def _load(self, **kwargs):
        # For vlan.interfacescollection.interface the partition is not valid
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        read_session = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']
        response = read_session.get(base_uri, **kwargs)
        self._local_update(response.json())
        self._build_meta_data_uri(self.selfLink)
        return self

    def load(self, **kwargs):
        self._load(**kwargs)
        return self

    def _check_load_parameters(self, **kwargs):
        '''Params given to load should at least satisfy required params.

        :params: kwargs
        :raises: MissingRequiredReadParameter
        '''
        key_set = set(kwargs.keys())
        required_minus_received =\
            self._meta_data['required_refresh_parameters'] - key_set
        if required_minus_received != set():
            error_message = 'Missing required params: %r'\
                % required_minus_received
            raise MissingRequiredReadParameter(error_message)

    def _update(self, **kwargs):
        update_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']
        read_only = self._meta_data.get('read_only_attributes', [])

        # Get the current state of the object on BigIP and check the generation
        # Use pop here because we don't want force in the data_dict
        force = self._check_force_arg(kwargs.pop('force', False))
        if not force:
            self._check_generation()

        # Save the meta data so we can add it back into self after we
        # load the new object.
        temp_meta = self.__dict__.pop('_meta_data')

        # Need to remove any of the Collection objects from self.__dict__
        # because these are sub-collections and _meta_data and
        # other non-BIGIP attrs are not removed from the sub-collections
        # See issue #146 for details
        for key, value in self.__dict__.items():
            if isinstance(value, Collection):
                self.__dict__.pop(key, '')
        data_dict = self.to_dict()

        # Remove any read-only attributes from our data_dict before we update
        # the data dict with the attributes.  If they pass in read-only attrs
        # in the method call we are going to let BIGIP let them know about it
        # when it fails
        for attr in read_only:
            data_dict.pop(attr, '')
        data_dict.update(kwargs)
        response = session.put(update_uri, json=data_dict)
        self._meta_data = temp_meta
        self._local_update(response.json())

    def update(self, **kwargs):
        # Need to implement checking for valid params here.
        self._update(**kwargs)

    def _delete(self, **kwargs):
        delete_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']

        # Check the generation for match before delete
        force = self._check_force_arg(kwargs.pop('force', True))
        if not force:
            self._check_generation()

        response = session.delete(delete_uri)
        if response.status_code == 200:
            self.__dict__ = {'deleted': True}

    def delete(self):
        # Need to implement checking for ? here.
        self._delete()
        # Need to implement correct teardown here.

    def _check_force_arg(self, force):
        if not isinstance(force, bool):
            raise InvalidForceType("force parameter must be type bool")
        return force

    def _check_generation(self):
        '''Check that the generation on the BigIP matches the object

        This will do a get to the objects URI and check that the generation
        returned in the JSON matches the one the object currently has.  If it
        does not it will raise the `GenerationMismatch` exception.
        '''
        session = self._meta_data['bigip']._meta_data['icr_session']
        response = session.get(self._meta_data['uri'])
        current_gen = response.json().get('generation', None)
        if current_gen is not None and current_gen != self.generation:
            error_message = ("The generation of the object on the BigIP " +
                             "(" + str(current_gen) + ")" +
                             " does not match the current object" +
                             "(" + str(self.generation) + ")")
            raise GenerationMismatch(error_message)



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip import exceptions
from f5.bigip.rest_collection import icontrol_rest_folder
from f5.bigip.rest_collection import log
from f5.common import constants as const
from f5.common.logger import Log

import json


class Route(object):
    def __init__(self, bigip):
        self.bigip = bigip
        self.domain_index = {'Common': 0}

    @icontrol_rest_folder
    @log
    def create(self, name=None, dest_ip_address=None, dest_mask=None,
               gw_ip_address=None, folder='Common'):
        """Create Route Entry """
        if dest_ip_address and dest_mask and gw_ip_address:
            folder = str(folder).replace('/', '')
            payload = dict()
            payload['name'] = name
            payload['partition'] = folder
            payload['gw'] = gw_ip_address
            payload['network'] = dest_ip_address + "/" + dest_mask
            request_url = self.bigip.icr_url + '/net/route/'
            response = self.bigip.icr_session.post(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 409:
                return True
            else:
                Log.error('route', response.text)
                raise exceptions.RouteCreationException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def delete(self, name=None, folder='Common'):
        """Delete Route Entry """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/route/'
        request_url += '~' + folder + '~' + name

        response = self.bigip.icr_session.delete(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            return True
        elif response.status_code == 404:
            return True
        else:
            Log.error('route', response.text)
            raise exceptions.RouteDeleteException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def delete_all(self, folder='Common'):
        """Delete Route Entries """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/route/'
        request_url += '?$select=name,selfLink'
        request_filter = 'partition eq ' + folder
        request_url += '&$filter=' + request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for item in response_obj['items']:
                    if item['name'].startswith(self.OBJ_PREFIX):
                        response = self.bigip.icr_session.delete(
                            self.bigip.icr_link(item['selfLink']),
                            timeout=const.CONNECTION_TIMEOUT)
                        if response.status_code > 400 and \
                           response.status_code != 404:
                            Log.error('route', response.text)
                            raise exceptions.RouteDeleteException(
                                response.text)
            return True
        else:
            Log.error('route', response.text)
            raise exceptions.RouteQueryException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def get_domain_by_id(self, folder='/Common', route_domain_id=0):
        """Get VLANs in Domain """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + \
            '/net/route-domain?$select=id,name,partition,vlans'
        if folder:
            request_filter = 'partition eq ' + folder
            request_url += '&$filter=' + request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for route_domain in response_obj['items']:
                    if int(route_domain['id']) == route_domain_id:
                        return route_domain
            return None
        else:
            if response.status_code != 404:
                Log.error('route-domain', response.text)
                raise exceptions.RouteQueryException(response.text)
        return None

    @icontrol_rest_folder
    @log
    def get_vlans_in_domain_by_id(self, folder='/Common', route_domain_id=0):
        """Get VLANs in Domain """
        route_domain = self.get_domain_by_id(
            folder=folder, route_domain_id=route_domain_id)
        vlans = []
        if 'vlans' in route_domain:
            for vlan in route_domain['vlans']:
                vlans.append(vlan)
        return vlans

    @icontrol_rest_folder
    @log
    def get_vlans_in_domain(self, folder='Common'):
        """Get VLANs in Domain """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + \
            '/net/route-domain?$select=name,partition,vlans'
        if folder:
            request_filter = 'partition eq ' + folder
            request_url += '&$filter=' + request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                vlans = []
                folder = str(folder).replace('/', '')
                for route_domain in response_obj['items']:
                    if route_domain['name'] == folder:
                        if 'vlans' in route_domain:
                            for vlan in route_domain['vlans']:
                                vlans.append(vlan)
                return vlans
            return []
        else:
            if response.status_code != 404:
                Log.error('route-domain', response.text)
                raise exceptions.RouteQueryException(response.text)
        return []

    @icontrol_rest_folder
    @log
    def add_vlan_to_domain_by_id(
            self, name=None, folder='Common', route_domain_id=0):
        """Add VLANs to Domain """
        folder = str(folder).replace('/', '')
        existing_vlans = self.get_vlans_in_domain_by_id(
            folder=folder, route_domain_id=route_domain_id)
        route_domain = self.get_domain_by_id(
            folder=folder, route_domain_id=route_domain_id)
        if not route_domain:
            raise exceptions.RouteUpdateException(
                "Cannot get route domain %s" % route_domain_id)
        if name not in existing_vlans:
            existing_vlans.append(name)
            vlans = dict()
            vlans['vlans'] = existing_vlans
            request_url = self.bigip.icr_url + '/net/route-domain/'
            request_url += '~' + folder + '~' + route_domain['name']
            response = self.bigip.icr_session.patch(
                request_url, data=json.dumps(vlans),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            else:
                Log.error('route-domain', response.text)
                raise exceptions.RouteUpdateException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def add_vlan_to_domain(self, name=None, folder='Common'):
        """Add VLANs to Domain """
        folder = str(folder).replace('/', '')
        existing_vlans = self.get_vlans_in_domain(folder=folder)
        if name not in existing_vlans:
            existing_vlans.append(name)
            vlans = dict()
            vlans['vlans'] = existing_vlans
            request_url = self.bigip.icr_url + '/net/route-domain/'
            request_url += '~' + folder + '~' + folder
            response = self.bigip.icr_session.patch(
                request_url, data=json.dumps(vlans),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            else:
                Log.error('route-domain', response.text)
                raise exceptions.RouteUpdateException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def remove_vlan_from_domain(self, name=None, folder='Common'):
        """Remove VLANs from Domain """
        folder = str(folder).replace('/', '')
        existing_vlans = self.get_vlans_in_domain(folder)
        if name in existing_vlans:
            existing_vlans.remove(name)
            vlans = dict()
            vlans['vlans'] = existing_vlans
            request_url = self.bigip.icr_url + '/net/route-domain/'
            request_url += '~' + folder + '~' + folder
            response = self.bigip.icr_session.patch(
                request_url, data=json.dumps(vlans),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            else:
                Log.error('route-domain', response.text)
                raise exceptions.RouteUpdateException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def create_domain(
            self, folder='Common', strict_route_isolation=False, is_aux=False):
        """Create route domain.

        is_aux: whether it is an auxiliary route domain beyond the main
                    route domain for the folder
        """

        folder = str(folder).replace('/', '')
        if not folder == 'Common':
            payload = dict()
            payload['partition'] = '/' + folder
            payload['id'] = self._get_next_domain_id()
            payload['name'] = folder
            if is_aux:
                payload['name'] += '_aux_' + str(payload['id'])
            if strict_route_isolation:
                payload['strict'] = 'enabled'
            else:
                payload['strict'] = 'disabled'
                payload['parent'] = '/Common/0'
            request_url = self.bigip.icr_url + '/net/route-domain/'
            response = self.bigip.icr_session.post(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return payload['id']
            elif response.status_code == 409:
                return True
            else:
                Log.error('route-domain', response.text)
                raise exceptions.RouteCreationException(response.text)
            return False
        return False

    @icontrol_rest_folder
    @log
    def delete_domain(self, folder='Common', name=None):
        """Delete route domain """
        folder = str(folder).replace('/', '')
        if not folder == 'Common':
            request_url = self.bigip.icr_url + '/net/route-domain/'
            if name:
                request_url += '~' + folder + '~' + name
            else:
                request_url += '~' + folder + '~' + folder
            response = self.bigip.icr_session.delete(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code != 404:
                Log.error('route-domain', response.text)
                raise exceptions.RouteDeleteException(response.text)
        return True

    @icontrol_rest_folder
    @log
    def domain_exists(self, folder='Common', route_domain_id=None):
        """Does the route domain exist?  """
        folder = str(folder).replace('/', '')
        if folder == 'Common':
            return True
        request_url = self.bigip.icr_url + '/net/route-domain/'
        if route_domain_id is None:
            request_url += '~' + folder + '~' + folder
        else:
            request_url += '~' + folder + '~' + \
                folder + '_aux_' + str(route_domain_id)
        request_url += '?$select=name'

        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code != 404:
            Log.error('route', response.text)
            raise exceptions.RouteQueryException(response.text)
        return False

    @icontrol_rest_folder
    @log
    def get_domain(self, folder='Common'):
        """Get route domain """
        folder = str(folder).replace('/', '')
        if folder == 'Common':
            return 0
        if folder in self.domain_index:
            return self.domain_index[folder]
        else:
            request_url = self.bigip.icr_url + '/net/route-domain/'
            request_url += '~' + folder + '~' + folder
            request_url += '?$select=id'
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                response_obj = json.loads(response.text)
                if 'id' in response_obj:
                    self.domain_index[folder] = int(response_obj['id'])
                    return self.domain_index[folder]
            elif response.status_code != 404:
                Log.error('route-domain', response.text)
                raise exceptions.RouteQueryException(response.text)
            return 0

    @icontrol_rest_folder
    @log
    def get_domain_ids(self, folder='Common'):
        """Get route domains """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/route-domain/'
        request_url += '?$select=id,partition'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                route_domains = []
                for route_domain in response_obj['items']:
                    if not folder:
                        route_domains.append(int(route_domain['id']))
                    elif route_domain['partition'] == folder:
                        route_domains.append(int(route_domain['id']))
                return route_domains
        elif response.status_code != 404:
            Log.error('route-domain', response.text)
            raise exceptions.RouteQueryException(response.text)
        return []

    @icontrol_rest_folder
    @log
    def get_domain_names(self, folder='Common'):
        """Get route domains """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/route-domain/'
        request_url += '?$select=partition,name'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                route_domains = []
                for route_domain in response_obj['items']:
                    if not folder:
                        route_domains.append(route_domain['name'])
                    elif route_domain['partition'] == folder:
                        route_domains.append(route_domain['name'])
                return route_domains
        elif response.status_code != 404:
            Log.error('route-domain', response.text)
            raise exceptions.RouteQueryException(response.text)
        return []

    @icontrol_rest_folder
    @log
    def exists(self, name=None, folder='Common'):
        """Does route exist? """
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/route/'
        request_url += '~' + folder + '~' + name
        request_url += '?$select=name'

        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        elif response.status_code != 404:
            Log.error('route', response.text)
            raise exceptions.RouteQueryException(response.text)
        return False

    def _get_next_domain_id(self):
        """Get next route domain id """
        request_url = self.bigip.icr_url + '/net/route-domain?$select=id'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        all_identifiers = []
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for route_domain in response_obj['items']:
                    all_identifiers.append(int(route_domain['id']))
                all_identifiers = sorted(all_identifiers)
                all_identifiers.remove(0)
        else:
            raise exceptions.RouteQueryException(response.text)

        lowest_available_index = 1
        for i in range(len(all_identifiers)):
            if all_identifiers[i] < lowest_available_index:
                if len(all_identifiers) > (i + 1):
                    if all_identifiers[i + 1] > lowest_available_index:
                        return lowest_available_index
                    else:
                        lowest_available_index = lowest_available_index + 1
            elif all_identifiers[i] == lowest_available_index:
                lowest_available_index = lowest_available_index + 1
        else:
            return lowest_available_index

    @log
    def set_strict_state(self, name=None, folder='Common', state='disabled'):
        """Route domain strict attribute """
        if name:
            folder = str(folder).replace('/', '')
            request_url = self.bigip.icr_url + '/net/route-domain/'
            request_url += '~' + folder + '~' + name
            payload = dict()
            payload['strict'] = state
            response = self.bigip.icr_session.patch(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            else:
                Log.error('route', response.text)
                raise exceptions.RouteUpdateException(response.text)
        return False



"""Manage application services on BIG-IP using REST interface """
# Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import Collection
from f5.bigip.resource import KindTypeMismatch
from f5.bigip.resource import Resource

from requests import HTTPError


class ApplicationCollection(Collection):
    def __init__(self, sys):
        super(ApplicationCollection, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            APLScriptCollection,
            CustomStatCollection,
            ServiceCollection,
            TemplateCollection
        ]


class APLScriptCollection(Collection):
    def __init__(self, application):
        super(APLScriptCollection, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [APLScript]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:apl-script:apl-scriptstate': APLScript}


class APLScript(Resource):
    def __init__(self, apl_script_collection):
        super(APLScript, self).__init__(apl_script_collection)
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:apl-script:apl-scriptstate'


class CustomStatCollection(Collection):
    def __init__(self, application):
        super(CustomStatCollection, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [CustomStat]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:custom-stat:custom-statstate': CustomStat}


class CustomStat(Resource):
    def __init__(self, custom_stat_collection):
        super(CustomStat, self).__init__(custom_stat_collection)
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:custom-stat:custom-statstate'


class ServiceCollection(Collection):
    def __init__(self, application):
        super(ServiceCollection, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [Service]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:service:servicestate': Service}


class Service(Resource):
    def __init__(self, service_collection):
        super(Service, self).__init__(service_collection)
        self._meta_data['required_creation_parameters'].update(('template',))
        self._meta_data['required_refresh_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:service:servicestate'

    def _create(self, **kwargs):
        '''Create service on device and create accompanying Python object.

        :params kwargs: keyword arguments passed in from create call
        :raises: KindTypeMismatch
        :raises: HTTPError
        :returns: Python Service object
        '''

        try:
            super(Service, self)._create(**kwargs)
        except HTTPError as ex:
            if "The configuration was updated successfully but could not be " \
                    "retrieved" not in ex.response.text:
                raise

            # BigIP will create in Common partition if none is given.
            # In order to create the uri properly in this class's load,
            # drop in Common as the partition in kwargs.
            if 'partition' not in kwargs:
                kwargs['partition'] = 'Common'
            # 'template' kwarg should not be used in the call to load becuase
            # the BigIP will return an error if it's present
            kwargs.pop('template')

            # If response was created successfully, do a local_update.
            # If not, call to overridden _load method via load
            self.load(**kwargs)
            if self.kind != self._meta_data['required_json_kind']:
                error_message = "For instances of type '%r' the corresponding"\
                    " kind must be '%r' but creation returned JSON with kind: %r"\
                    % (self.__class__.__name__,
                       self._meta_data['required_json_kind'],
                       self.kind)
                raise KindTypeMismatch(error_message)

        return self

    def update(self, **kwargs):
        '''Push local updates to the object on the device.

        :params kwargs: keyword arguments for accessing/modifying the object
        :returns: updated Python object
        '''

        inherit_device_group = self.__dict__.get('inheritedDevicegroup', False)
        if inherit_device_group == 'true':
            self.__dict__.pop('deviceGroup')
        return self._update(**kwargs)

    def _load(self, **kwargs):
        '''Load python Service object with response JSON from BigIP.

        :params kwargs: keyword arguments for talking to the device
        :returns: populated Service object
        '''

        self._check_load_parameters(**kwargs)
        name = kwargs.pop('name')
        partition = kwargs.pop('partition')
        read_session = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']

        name = name.replace('/', '~')
        load_uri = '%s~%s~%s.app~%s' % (base_uri, partition, name, name)

        response = read_session.get(load_uri, uri_as_parts=False, **kwargs)
        self._local_update(response.json())
        self._build_meta_data_uri(self.selfLink)
        return self


class TemplateCollection(Collection):
    def __init__(self, application):
        super(TemplateCollection, self).__init__(application)
        self._meta_data['allowed_lazy_attributes'] = [Template]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:application:template:templatestate': Template}


class Template(Resource):
    def __init__(self, template_collection):
        super(Template, self).__init__(template_collection)
        self._meta_data['required_json_kind'] =\
            'tm:sys:application:template:templatestate'



""" Classes and functions for configuring ARP on bigip """
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=broad-except,no-self-use

from f5.bigip import exceptions
from f5.bigip.rest_collection import icontrol_folder
from f5.bigip.rest_collection import icontrol_rest_folder
from f5.bigip.rest_collection import log
from f5.common import constants as const
from f5.common.logger import Log

import json
import netaddr
import urllib


class ARP(object):
    """Class for configuring ARP on bigip """

    def __init__(self, bigip):
        self.bigip = bigip
        # add iControl interfaces if they don't exist yet
        self.bigip.icontrol.add_interfaces(['Networking.ARP'])

        # iControl helper objects
        self.net_arp = self.bigip.icontrol.Networking.ARP

    # pylint: disable=pointless-string-statement
    '''
    @icontrol_rest_folder
    @log
    def create(self, ip_address=None, mac_address=None, folder='Common'):
        payload = dict()
        payload['name'] = ip_address
        payload['partition'] = folder
        payload['ipAddress'] = ip_address
        payload['macAddress'] = mac_address
        request_url = self.icr_url + '/net/arp/'
        response = self.icr_session.post(request_url,
                              data=json.dumps(payload),
                              timeout=)
        Log.debug('ARP::create response',
                  '%s' % response.json())
        if response.status_code < 400:
            return True
        elif response.status_code == 409:
            return True
        else:
            raise exceptions.StaticARPCreationException(response.text)
    '''
    # pylint: enable=pointless-string-statement

    @icontrol_folder
    @log
    def create(self, ip_address=None, mac_address=None, folder='Common'):
        """Create an ARP static entry """
        if not self.exists(ip_address=ip_address, folder=folder):
            # ARP entries can't handle %0 on them like other
            # TMOS objects.
            ip_address = self._remove_route_domain_zero(ip_address)
            try:
                create_arp = self.net_arp.typefactory.create
                entry = create_arp('Networking.ARP.StaticEntry')
                entry.address = ip_address
                entry.mac_address = mac_address
                self.net_arp.add_static_entry([entry])
                return True
            except Exception as exc:
                Log.error('ARP', 'create exception: ' + exc.message)
                raise exceptions.StaticARPCreationException(exc.message)
        return False

    # pylint: disable=pointless-string-statement
    '''
    @icontrol_rest_folder
    def delete(self, ip_address=None, folder='Common'):
        if ip_address:
            folder = str(folder).replace('/', '')
            request_url = self.bigip.icr_url + '/net/arp/'
            request_url += '~' + folder + '~' + urllib.quote(
                                  self._remove_route_domain_zero(ip_address))
            response = self.bigip.icr_session.delete(request_url,
                                            timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                return True
            elif response.status_code == 404:
                return True
            else:
                raise exceptions.StaticARPDeleteException(response.text)
                Log.error('ARP', response.text)
        return False
    '''
    # pylint: enable=pointless-string-statement

    @icontrol_folder
    @log
    def delete(self, ip_address=None, folder='Common'):
        """Delete an ARP static entry """
        if self.exists(ip_address=ip_address, folder=folder):
            # ARP entries can't handle %0 on them like other
            # TMOS objects.
            ip_address = self._remove_route_domain_zero(ip_address)
            try:
                self.net_arp.delete_static_entry_v2(
                    ['/' + folder + '/' + ip_address])
                return True
            except Exception as exc:
                Log.error('ARP', 'delete exception: ' + exc.message)
                raise exceptions.StaticARPDeleteException(exc.message)
        return False

    @icontrol_folder
    @log
    def delete_by_mac(self, mac_address=None, folder='Common'):
        """Delete an ARP static entry by MAC address """
        if mac_address:
            arps = self.get_arps(None, folder)
            for arp in arps:
                for ip_address in arp:
                    if arp[ip_address] == mac_address:
                        self.delete(ip_address=ip_address, folder=folder)

    @icontrol_folder
    @log
    def delete_by_subnet(self, subnet=None, mask=None, folder='Common'):
        """Delete ARP static entries on subnet """
        if subnet:
            mask_div = subnet.find('/')
            if mask_div > 0:
                try:
                    rd_div = subnet.find('%')
                    if rd_div > -1:
                        network = netaddr.IPNetwork(
                            subnet[0:mask_div][0:rd_div] + subnet[mask_div:])
                    else:
                        network = netaddr.IPNetwork(subnet)
                except Exception as exc:
                    Log.error('ARP', exc.message)
                    return []
            elif not mask:
                return []
            else:
                try:
                    rd_div = subnet.find('%')
                    if rd_div > -1:
                        network = netaddr.IPNetwork(
                            subnet[0:rd_div] + '/' + mask)
                    else:
                        network = netaddr.IPNetwork(subnet + '/' + mask)
                except Exception as exc:
                    Log.error('ARP', exc.message)
                    return []

            return self._delete_by_network(folder, network)

    def _delete_by_network(self, folder, network):
        """Delete for network """
        if not network:
            return []
        mac_addresses = []
        request_url = self.bigip.icr_url + '/net/arp'
        request_filter = 'partition eq ' + folder
        request_url += '?$filter=' + request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for arp in response_obj['items']:
                    ad_rd_div = arp['ipAddress'].find('%')
                    if ad_rd_div > -1:
                        address = netaddr.IPAddress(
                            arp['ipAddress'][0:ad_rd_div])
                    else:
                        address = netaddr.IPAddress(arp['ipAddress'])

                    if address in network:
                        mac_addresses.append(arp['macAddress'])
                        self.delete(arp['ipAddress'],
                                    folder=arp['partition'])
        return mac_addresses

    @icontrol_rest_folder
    @log
    def get_arps(self, ip_address=None, folder='Common'):
        """Get ARP static entry """
        folder = str(folder).replace('/', '')
        if ip_address:
            request_url = self.bigip.icr_url + '/net/arp/'
            request_url += '~' + folder + '~' + urllib.quote(
                self._remove_route_domain_zero(ip_address))
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            Log.debug('ARP::get response',
                      '%s' % response.json())
            if response.status_code < 400:
                response_obj = json.loads(response.text)
                return [
                    {response_obj['name']:
                     response_obj['macAddress']}
                ]
            else:
                Log.error('ARP', response.text)
                raise exceptions.StaticARPQueryException(response.text)
        else:
            request_url = self.bigip.icr_url + '/net/arp'
            request_filter = 'partition eq ' + folder
            request_url += '?$filter=' + request_filter
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            Log.debug('ARP::get response',
                      '%s' % response.json())
            if response.status_code < 400:
                response_obj = json.loads(response.text)
                if 'items' in response_obj:
                    arps = []
                    for arp in response_obj['items']:
                        arps.append(
                            {arp['name']:
                             arp['macAddress']}
                        )
                    return arps
            else:
                Log.error('ARP', response.text)
                raise exceptions.StaticARPQueryException(response.text)
        return []

    # pylint: disable=pointless-string-statement
    '''
    @icontrol_rest_folder
    def delete_all(self, folder='Common'):
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/arp/'
        request_url += '?$select=name,selfLink'
        request_filter = 'partition eq ' + folder
        request_url += '&$filter=' + request_filter
        response = self.bigip.icr_session.get(request_url,
                                             timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            deletions = []
            if 'items' in response_obj:
                for item in response_obj['items']:
                    if item['name'].startswith(self.OBJ_PREFIX):
                        deletions.append(self.bigip.icr_link(item['selfLink']))
            for delete in deletions:
                response = self.bigip.icr_session.delete(delete,
                                              timeout=const.CONNECTION_TIMEOUT)
                if response.status_code > 400 and \
                  (not response.status_code == 404):
                    Log.error('ARP', response.text)
                    return False
        elif response.status_code == 404:
            return True
        else:
            Log.error('ARP', response.text)
            exceptions.StaticARPDeleteException(response.text)
     '''
    # pylint: enable=pointless-string-statement

    @icontrol_folder
    @log
    def delete_all(self, folder='Common'):
        """Delete all ARP entries """
        try:
            self.net_arp.delete_all_static_entries()
        except Exception as exc:
            Log.error('ARP', 'delete exception: ' + exc.message)
            raise exceptions.StaticARPDeleteException(exc.message)

    # pylint: disable=pointless-string-statement
    '''
    @icontrol_rest_folder
    def exists(self, ip_address=None, folder='Common'):
        folder = str(folder).replace('/', '')
        request_url = self.bigip.icr_url + '/net/arp/'
        request_url += '~' + folder + '~' + urllib.quote(
                        self._remove_route_domain_zero(ip_address))
        response = self.bigip.icr_session.get(request_url,
                                  timeout=const.CONNECTION_TIMEOUT)
        Log.debug('ARP::exists response',
                      '%s' % response.text)
        if response.status_code < 400:
            return True
        return False
    '''
    # pylint: enable=pointless-string-statement

    @icontrol_folder
    @log
    def exists(self, ip_address=None, folder='Common'):
        """Does ARP entry exist? """
        # ARP entries can't handle %0 on them like other
        # TMOS objects.
        ip_address = self._remove_route_domain_zero(ip_address)
        try:
            arp_list = self.net_arp.get_static_entry_list()
        except Exception as exc:
            Log.error('ARP', 'query exception: %s on %s' %
                      (exc.message, self.bigip.device_name))
            raise exceptions.StaticARPQueryException(exc.message)

        if '/' + folder + '/' + ip_address in arp_list:
            return True
        else:
            return False

    def _remove_route_domain_zero(self, ip_address):
        """Remove route domain zero from ip_address """
        decorator_index = ip_address.find('%0')
        if decorator_index > 0:
            ip_address = ip_address[:decorator_index]
        return ip_address



# Copyright 2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.cm.cluster import Cluster
from f5.bigip.cm.device import Device

base_uri = 'cm/'


class CM(object):
    def __init__(self, bigip):
        self.interfaces = {}
        self.bigip = bigip

    @property
    def cluster(self):
        if 'cluster' in self.interfaces:
            return self.interfaces['cluster']
        else:
            cluster = Cluster(self.bigip)
            self.interfaces['cluster'] = cluster
            return cluster

    @property
    def device(self):
        if 'device' in self.interfaces:
            return self.interfaces['device']
        else:
            device = Device(self.bigip)
            self.interfaces['device'] = device
            return device



"""Classes and functions for configuring bigip system """
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=broad-except

from f5.bigip import exceptions
from f5.bigip.rest_collection import log
from f5.common import constants as const
from f5.common.logger import Log

from suds import WebFault

import json
import time
import uuid


class System(object):
    """Class for configuring bigip system """

    OBJ_PREFIX = 'uuid_'

    def __init__(self, bigip):
        self.bigip = bigip

        self.bigip.icontrol.add_interfaces(['Management.Folder',
                                            'System.Session',
                                            'System.SystemInfo',
                                            'System.SoftwareManagement',
                                            'System.ConfigSync',
                                            'System.VCMP']
                                           )

        # iControl helper objects
        self.mgmt_folder = self.bigip.icontrol.Management.Folder
        self.sys_session = self.bigip.icontrol.System.Session
        self.sys_info = self.bigip.icontrol.System.SystemInfo
        self.sys_swmgmt = self.bigip.icontrol.System.SoftwareManagement
        self.sys_config_sync = self.bigip.icontrol.System.ConfigSync
        self.sys_vcmp = self.bigip.icontrol.System.VCMP

        # create stubs to hold static system params to avoid redundant calls
        self.version = None
        self.current_folder = None
        self.systeminfo = None
        self.exempt_folders = ['/', 'Common']
        self.existing_folders = {}
        self.existint_folders_updated = None

    @log
    def folder_exists(self, folder):
        """Does folder exist? """
        if folder:
            folder = str(folder).replace('/', '')
            if folder == 'Common':
                return True
            if folder in self.existing_folders:
                now = time.time()
                if self.existint_folders_updated:
                    if ((now - self.existint_folders_updated) <
                            const.FOLDER_CACHE_TIMEOUT):
                        return True
                    else:
                        del self.existing_folders[folder]
            request_url = self.bigip.icr_url + '/sys/folder/'
            request_url += '~' + folder
            request_url += '?$select=name'
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                self.existing_folders[folder] = 1
                return True
            elif response.status_code == 404:
                return False
            else:
                Log.error('folder', response.text)
                raise exceptions.SystemQueryException(response.text)
        return False

    @log
    def create_folder(self, folder, change_to=False, traffic_group=None):
        """Create folder """
        if folder:
            folder = str(folder).replace('/', '')
            request_url = self.bigip.icr_url + '/sys/folder/'
            payload = dict()
            payload['name'] = folder
            payload['subPath'] = '/'
            payload['fullPath'] = '/' + folder
            payload['hidden'] = False
            payload['inheritedDevicegroup'] = True
            if traffic_group:
                payload['trafficGroup'] = traffic_group
                payload['inheritedTrafficGroup'] = False
            else:
                payload['inheritedTrafficGroup'] = True

            response = self.bigip.icr_session.post(
                request_url, data=json.dumps(payload),
                timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                if change_to:
                    self.existing_folders[folder] = 1
                    self.set_folder(folder)
                else:
                    self.set_folder('/Common')
                return True
            else:
                Log.error('folder', response.text)
                raise exceptions.SystemCreationException(response.text)
        return False

    def force_root_folder(self):
        """Force iControl SOAP context into root folder.

            This is typically done before deleting a folder.
            We need to ensure the internal context of the iControl
            SOAP portal is not associated with a folder before
            the folder is deleted or the SOAP portal will become
            inoperative.
            We need to do a fake query and fake command
            because setting your active folder, by itself, does
            not do anything.
        """
        self.sys_session.set_active_folder('/')
        self.current_folder = '/'
        self.mgmt_folder.get_list()
        fakename = '/set-folder-workaround-' + str(uuid.uuid4())[0:8]
        try:
            self.mgmt_folder.delete_folder([fakename])
        except WebFault:
            pass

    @log
    def delete_folder(self, folder):
        """Delete folder """
        if folder:
            folder = str(folder).replace('/', '')
            request_url = self.bigip.icr_url + '/sys/folder/~' + folder
            response = self.bigip.icr_session.delete(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                if folder in self.existing_folders:
                    del self.existing_folders[folder]
                self.set_folder('/Common')
                return True
            elif response.status_code == 404:
                return True
            else:
                Log.error('folder', response.text)
                raise exceptions.SystemDeleteException(response.text)
        return False

    @log
    def get_folders(self):
        """Get Folders """
        request_url = self.bigip.icr_url + '/sys/folder/'
        request_url += '?$select=name'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        return_list = []
        if response.status_code < 400:
            return_obj = json.loads(response.text)
            if 'items' in return_obj:
                for folder in return_obj['items']:
                    return_list.append(folder['name'])
        elif response.status_code != 404:
            Log.error('folder', response.text)
            raise exceptions.SystemQueryException(response.text)
        return return_list

    @log
    def set_folder(self, folder):
        """Set Folder """
        if not folder:
            msg = 'set_folder failed: No folder specified!'
            Log.error('System', msg)
            raise exceptions.SystemUpdateException(msg)

        if not self.folder_exists(folder):
            msg = 'set_folder:set_active_folder failed, ' + \
                  'folder does not exist!'
            Log.error('System', msg)
            raise exceptions.SystemUpdateException(msg)

        if not str(folder).startswith('/'):
            folder = '/' + folder
        if self.current_folder and folder == self.current_folder:
            return
        try:
            self.sys_session.set_active_folder(folder)
            self.current_folder = folder
        except WebFault as webfault:
            Log.error('System',
                      'set_folder:set_active_folder failed: ' +
                      str(webfault.message))
            raise exceptions.SystemUpdateException(webfault.message)

    @log
    def purge_folder_contents(self, folder, bigip=None):
        """Purge Folder of contents """
        if not bigip:
            bigip = self.bigip
        if folder not in self.exempt_folders:
            bigip.virtual_server.delete_all(folder=folder)
            bigip.pool.delete_all(folder=folder)
            bigip.monitor.delete_all(folder=folder)
            bigip.snat.delete_all(folder=folder)
            bigip.virtual_server.delete_all_persistence_profiles(folder=folder)
            bigip.virtual_server.delete_all_http_profiles(folder=folder)
            bigip.rule.delete_all(folder=folder)
            bigip.arp.delete_all(folder=folder)
            bigip.selfip.delete_all(folder=folder)
            bigip.vlan.delete_all(folder=folder)
            bigip.l2gre.delete_all(folder=folder)
            bigip.route.delete_domain(folder=folder)
        else:
            Log.error('folder',
                      'Request to purge exempt folder %s ignored.' % folder)

    @log
    def purge_folder(self, folder, bigip=None):
        """Purge folder """
        if not bigip:
            bigip = self.bigip
        if folder not in self.exempt_folders:
            bigip.system.delete_folder(bigip.decorate_folder(folder))
        else:
            Log.error('folder',
                      'Request to purge exempt folder %s ignored.' % folder)

    @log
    def purge_orphaned_folders_contents(self, known_folders, bigip=None):
        """Purge Folder of contents """
        if not bigip:
            bigip = self.bigip
        existing_folders = bigip.system.get_folders()
        # remove all folders which are default
        existing_folders.remove('/')
        existing_folders.remove('Common')
        # remove all folders which are not managed
        # with this object prefix.
        # copy the list so we are not modifying it while traversing it.
        existing_folders_copy = list(existing_folders)
        for folder in existing_folders_copy:
            if not folder.startswith(self.OBJ_PREFIX):
                existing_folders.remove(folder)
            # iapp folders need to be purged by removing the iapp
            if folder.endswith('.app'):
                existing_folders.remove(folder)
        for folder in known_folders:
            decorated_folder = bigip.decorate_folder(folder)
            if decorated_folder in existing_folders:
                existing_folders.remove(decorated_folder)
        # anything left should be purged
        if existing_folders:
            Log.debug('system',
                      'purging orphaned folders contents: %s'
                      % existing_folders)
        for folder in existing_folders:
            try:
                bigip.system.purge_folder_contents(folder, bigip)
            except Exception as exc:
                Log.error('purge_orphaned_folders_contents', exc.message)

    @log
    def purge_orphaned_folders(self, known_folders, bigip=None):
        """Purge Folders """
        if not bigip:
            bigip = self.bigip
        existing_folders = bigip.system.get_folders()
        # remove all folders which are default
        existing_folders.remove('/')
        existing_folders.remove('Common')
        # remove all folders which are not managed
        # with this object prefix
        existing_folders_copy = list(existing_folders)
        for folder in existing_folders_copy:
            if not folder.startswith(self.OBJ_PREFIX):
                existing_folders.remove(folder)
            # iapp folders need to be purged by removing the iapp
            if folder.endswith('.app'):
                existing_folders.remove(folder)
        for folder in known_folders:
            decorated_folder = bigip.decorate_folder(folder)
            if decorated_folder in existing_folders:
                existing_folders.remove(decorated_folder)
        # anything left should be purged
        if existing_folders:
            Log.debug('system', 'purging orphaned folders: %s'
                      % existing_folders)
        for folder in existing_folders:
            try:
                bigip.system.purge_folder(folder, bigip)
            except Exception as exc:
                Log.error('purge_orphaned_folders', exc.message)

    @log
    def purge_all_folders(self, bigip=None):
        """Purge all folders """
        if not bigip:
            bigip = self.bigip
        existing_folders = bigip.system.get_folders()
        for folder in existing_folders:
            if folder.startswith(bigip.system.OBJ_PREFIX):
                bigip.system.purge_folder(folder)

    @log
    def get_hostname(self):
        """Get bigip hostname """
        request_url = self.bigip.icr_url + \
            '/sys/global-settings?$select=hostname'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            return response_obj['hostname']
        else:
            raise exceptions.SystemQueryException(response.text)

    @log
    def set_hostname(self, hostname):
        """Set bigip hostname """
        request_url = self.bigip.icr_url + '/sys/global-settings'
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps({'hostname': hostname}),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            raise exceptions.SystemUpdateException(response.text)

    @log
    def get_ntp_server(self):
        """Get bigip ntp server """
        request_url = self.bigip.icr_url + \
            '/sys/ntp?$select=servers'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'servers' in response_obj:
                return response_obj['servers'][0]
            else:
                return None
        else:
            raise exceptions.SystemQueryException(response.text)

    @log
    def set_ntp_server(self, addr):
        """Set bigip ntp server """
        request_url = self.bigip.icr_url + '/sys/ntp'
        if not isinstance(addr, list):
            addr = [addr]
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps({'servers': addr}),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            raise exceptions.SystemUpdateException(response.text)

    @log
    def get_active_modules(self):
        """Get bigip active modules """
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '?$select=activeModules,selfDevice'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for device in response_obj['items']:
                    if device['selfDevice']:
                        return device['activeModules']
        else:
            raise exceptions.SystemQueryException(response.text)

    @log
    def get_platform(self):
        """Get platform """
        if not self.systeminfo:
            try:
                self.systeminfo = self.sys_info.get_system_information()
            except Exception as exc:
                raise exceptions.SystemQueryException(exc.message)
        return self.systeminfo.product_category

    @log
    def get_serial_number(self):
        """Get serial number """
        if not self.systeminfo:
            try:
                self.systeminfo = self.sys_info.get_system_information()
            except Exception as exc:
                raise exceptions.SystemQueryException(exc.message)
        return self.systeminfo.chassis_serial

    @log
    def get_version(self):
        """Get version """
        if not self.version:
            try:
                self.version = self.sys_info.get_version()
            except Exception as exc:
                raise exceptions.SystemQueryException(exc.message)
        return self.version

    @log
    def get_major_version(self):
        """Get major version """
        return self.get_version().split('_v')[1].split('.')[0]

    @log
    def get_minor_version(self):
        """Get minor version """
        return self.get_version().split('_v')[1].split('.')[1]

    @log
    def get_license_operational(self):
        """Get license operational """
        request_url = self.bigip.icr_url + '/sys/db/license.operational'
        request_url += '?$select=value'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'value' in response_obj:
                return response_obj['value'] == 'true'
            return None
        elif response.status_code == 404:
            return None
        else:
            raise exceptions.SystemQueryException(response.text)

    @log
    def get_provision_extramb(self):
        """Get provisioned extramb for large management memory """
        request_url = self.bigip.icr_url + '/sys/db/provision.extramb'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'value' in response_obj:
                return response_obj['value']
            return 0
        else:
            raise exceptions.SystemQueryException(response.text)

    @log
    def set_provision_extramb(self, extramdb=500):
        """Set provisioned extramb for large management memory """
        request_url = self.bigip.icr_url + '/sys/db/provision.extramb'
        response = self.bigip.icr_session.put(
            request_url, data=json.dumps({'value': extramdb}),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            raise exceptions.SystemUpdateException(response.text)

    @log
    def get_tunnel_sync(self):
        """Get option for syncing tunnels """
        request_url = self.bigip.icr_url + '/sys/db/iptunnel.configsync'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)

        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'value' in response_obj:
                return response_obj['value']
            return None
        elif response.status_code != 404:
            raise exceptions.SystemQueryException(response.text)

    @log
    def set_tunnel_sync(self, enabled=False):
        """Set option for syncing tunnels """
        request_url = self.bigip.icr_url + '/sys/db/iptunnel.configsync'
        if enabled:
            response = self.bigip.icr_session.put(
                request_url, data=json.dumps({'value': 'enable'}),
                timeout=const.CONNECTION_TIMEOUT)
        else:
            response = self.bigip.icr_session.put(
                request_url, data=json.dumps({'value': 'disable'}),
                timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'value' in response_obj:
                return response_obj['value']
            return None
        elif response.status_code != 404:
            raise exceptions.SystemUpdateException(response.text)

    @log
    def force_to_standby(self, traffic_group):
        """Force traffic group to standby """
        request_url = self.bigip.icr_url + '/sys/failover'

        payload = dict()
        payload['command'] = 'run'
        payload['standby'] = None
        payload['trafficGroup'] = traffic_group

        response = self.bigip.icr_session.post(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True

        Log.error('traffic-group', response.text)
        raise exceptions.SystemUpdateException(response.text)



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import json
import time

from f5.bigip import exceptions
from f5.bigip.rest_collection import log
from f5.common import constants as const
from f5.common.logger import Log


# Management - Device
class Device(object):
    def __init__(self, bigip):
        self.bigip = bigip

        self.bigip.icontrol.add_interfaces(['Management.Trust'])
        self.mgmt_trust = self.bigip.icontrol.Management.Trust

        # create empty lock instance ID
        self.lock = None
        self.devicename = None

    @log
    def get_device_name(self):
        """Get device name """
        if not self.devicename:
            request_url = self.bigip.icr_url + '/cm/device'
            request_filter = '/?$select=name,selfDevice'
            request_filter += '&filter partition eq Common'
            request_url += request_filter
            response = self.bigip.icr_session.get(
                request_url, timeout=const.CONNECTION_TIMEOUT)
            if response.status_code < 400:
                response_obj = json.loads(response.text)
                if 'items' in response_obj:
                    devices = response_obj['items']
                    for device in devices:
                        if device['selfDevice'] == 'true':
                            self.devicename = device['name']
            else:
                Log.error('device', response.text)
                raise exceptions.DeviceQueryException(response.text)
        return self.devicename

    @log
    def get_all_device_names(self):
        """Get all device name """
        request_url = self.bigip.icr_url + '/cm/device'
        request_filter = '/?$select=name&filter partition eq Common'
        request_url += request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                devices = response_obj['items']
                device_names = []
                for device in devices:
                    device_names.append(device['name'])
                return device_names
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return []

    @log
    def get_lock(self):
        """Get device lock """
        current_lock = self._get_lock()
        new_lock = int(time.time())

        if current_lock:
            if (new_lock - current_lock) > const.CONNECTION_TIMEOUT:
                Log.info('Device', 'Locking device %s with lock %s'
                         % (self.get_device_name(), new_lock))
                self._set_lock(new_lock)
                return True
            else:
                return False
        else:
            Log.info('Device', 'Locking device %s with lock %s'
                     % (self.get_device_name(), new_lock))
            self._set_lock(int(time.time()))
            return True

    @log
    def release_lock(self):
        """Release device lock """
        current_lock = self._get_lock()

        if current_lock == self.lock:
            Log.info('Device', 'Releasing device lock for %s'
                     % self.get_device_name())
            self._set_lock(None)
            return True
        else:
            Log.info('Device', 'Device has foreign lock instance on %s '
                     % self.get_device_name() + ' with lock %s '
                     % current_lock)
            return False

    def _get_lock(self):
        """Get device lock """
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '?$select=selfDevice,comment'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        current_lock = ''
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                devices = response_obj['items']
                for device in devices:
                    if device['selfDevice']:
                        if 'comment' in device:
                            current_lock = device['comment']
        if current_lock.startswith(const.DEVICE_LOCK_PREFIX):
            return int(current_lock.replace(const.DEVICE_LOCK_PREFIX, ''))

    def _set_lock(self, lock):
        """Set device lock """
        dev_name = self.get_device_name()
        if lock:
            self.lock = lock
            lock_comment = const.DEVICE_LOCK_PREFIX + str(lock)
        else:
            lock_comment = ''
        request_url = self.bigip.icr_url + '/cm/device/'
        request_url += '~Common~' + dev_name
        payload = dict()
        payload['comment'] = lock_comment
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        return False

    @log
    def get_mgmt_addr(self):
        """Get device management ip """
        request_url = self.bigip.icr_url + '/cm/device/~Common'
        request_url += '~' + self.get_device_name()
        request_filter = '/?$select=managementIp'
        request_url += request_filter
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            return response_obj['managementIp']
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return None

    @log
    def get_all_mgmt_addrs(self):
        """Get device management ips """
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '/?$select=managementIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                mgmt_addrs = []
                for device in response_obj['items']:
                    mgmt_addrs.append(device['managementIp'])
            return mgmt_addrs
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return None

    @log
    def get_mgmt_addr_by_device(self, devicename):
        request_url = self.bigip.icr_url + '/cm/device'
        request_url += '/?$select=managementIp,name'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for device in response_obj['items']:
                    if device['name'] == devicename:
                        return device['managementIp']
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return None

    @log
    def get_configsync_addr(self):
        """Get device config sync ip """
        request_url = self.bigip.icr_url + '/cm/device/~Common'
        request_url += '~' + self.get_device_name()
        request_url += '/?$select=configsyncIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            return response_obj['configsyncIp']
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
            return None

    @log
    def set_configsync_addr(self, ip_address=None, folder='/Common'):
        """Set device config sync ip """
        dev_name = self.get_device_name()
        request_url = self.bigip.icr_url + '/cm/device/'
        request_url += '~Common~' + dev_name
        payload = dict()
        if not ip_address:
            payload['configsyncIp'] = None
        else:
            payload['configsyncIp'] = ip_address
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return False

    @log
    def get_primary_mirror_addr(self):
        """Get device primary mirror ip """
        request_url = self.bigip.icr_url + '/cm/device/~Common'
        request_url += '~' + self.get_device_name()
        request_url += '/?$select=mirrorIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            primary = response_obj['mirrorIp']
            if primary == 'any6':
                return None
            else:
                return primary
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return None

    @log
    def get_secondary_mirror_addr(self):
        """Get device secondary mirror ip """
        request_url = self.bigip.icr_url + '/cm/device/~Common'
        request_url += '~' + self.get_device_name()
        request_url += '/?$select=mirrorSecondaryIp'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            secondary = response_obj['mirrorSecondaryIp']
            if secondary == 'any6':
                return None
            else:
                return secondary
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return None

    @log
    def set_primary_mirror_addr(self, ip_address=None, folder='/Common'):
        """Set device primary mirror ip """
        dev_name = self.get_device_name()
        request_url = self.bigip.icr_url + '/cm/device/'
        request_url += '~Common~' + dev_name
        payload = dict()
        if not ip_address:
            payload['mirrorIp'] = '::'
        else:
            payload['mirrorIp'] = ip_address
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceUpdateException(response.text)
        return False

    @log
    def set_secondary_mirror_addr(self, ip_address=None, folder='/Common'):
        """Set device secondary mirror ip """
        dev_name = self.get_device_name()
        request_url = self.bigip.icr_url + '/cm/device/'
        request_url += '~Common~' + dev_name
        payload = dict()
        if not ip_address:
            payload['mirrorSecondaryIp'] = '::'
        else:
            payload['mirrorSecondaryIp'] = ip_address
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceUpdateException(response.text)
        return False

    @log
    def get_failover_addrs(self):
        """Get device failover ips """
        request_url = self.bigip.icr_url + '/cm/device/~Common'
        request_url += '~' + self.get_device_name()
        request_url += '/?$select=unicastAddress'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            return_addresses = []
            if 'unicastAddress' in response_obj:
                uas = response_obj['unicastAddress']
                for ua in uas:
                    return_addresses.append(ua['ip'])
                return return_addresses
            else:
                return []
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return []

    @log
    def set_failover_addrs(self, ip_addrs=None, folder='/Common'):
        """Get device failover ips """
        dev_name = self.get_device_name()
        dev_ip = self.get_mgmt_addr()
        request_url = self.bigip.icr_url + '/cm/device/'
        request_url += '~Common~' + dev_name
        payload = dict()
        unicast_addresses = []
        if len(ip_addrs):
            unicast_addresses.append({'effectiveIp': dev_ip,
                                      'effectivePort': 1026,
                                      'ip': dev_ip,
                                      'port': 1026})
            for ip_address in ip_addrs:
                unicast_addresses.append({'effectiveIp': ip_address,
                                          'effectivePort': 1026,
                                          'ip': ip_address,
                                          'port': 1026})
        payload['unicastAddress'] = unicast_addresses
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceUpdateException(response.text)
        return False

    @log
    def get_failover_state(self):
        """Get device failover state """
        request_url = self.bigip.icr_url + '/cm/device/~Common'
        request_url += '~' + self.get_device_name()
        request_url += '/?$select=failoverState'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            return response_obj['failoverState']
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return None

    @log
    def get_device_group(self):
        """Get device group """
        request_url = self.bigip.icr_url + '/cm/device-group'
        request_url += '/?$select=name,type'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                dsgs = response_obj['items']
                for dsg in dsgs:
                    if dsg['type'] == 'sync-failover':
                        return dsg['name']
                return None
        elif response.status_code == 404:
            return None
        else:
            Log.error('device-group', response.text)
            raise exceptions.DeviceQueryException(response.text)
        return None

    @log
    def remove_from_device_group(self, name=None, folder='/Common'):
        """Remove device from group """
        device_group = self.get_device_group()
        if device_group:
            return self.bigip.cluster.remove_devices(device_group,
                                                     [self.get_device_name()])

    @log
    def remove_all_peers(self):
        """Remove all peers from group """
        self.bigip.system.set_folder('/Common')
        current_dev_name = self.get_device_name()
        devs_to_remove = []
        for dev in self.get_all_device_names():
            if dev != current_dev_name:
                devs_to_remove.append(dev)
        if devs_to_remove:
            try:
                self.mgmt_trust.remove_device(devs_to_remove)
            except Exception as e:
                Log.error('device', e.message)
                raise exceptions.DeviceUpdateException(e.message)
        try:
            self.remove_metadata(
                None, {'root_device_name': None,
                       'root_device_mgmt_address': None})
        except exceptions.DeviceUpdateException:
            pass

    @log
    def reset_trust(self, new_name):
        """Remove trust """
        self.bigip.system.set_folder('/Common')
        self.remove_all_peers()
        try:
            self.mgmt_trust.reset_all(new_name, False, '', '')
        except Exception as e:
            Log.error('device', e.message)
            raise exceptions.DeviceUpdateException(e.message)
        try:
            self.remove_metadata(
                None, {'root_device_name': None,
                       'root_device_mgmt_address': None})
        except exceptions.DeviceUpdateException:
            pass

        self.devicename = None
        self.get_device_name()

    @log
    def set_metadata(self, name=None, device_dict=None):
        """Set device metadata """
        if not name:
            name = self.get_device_name()
        if isinstance(device_dict, dict):
            str_comment = json.dumps(device_dict)
        else:
            str_comment = device_dict

        request_url = self.bigip.icr_url + '/cm/device/~Common~'
        request_url += name
        payload = dict()
        payload['description'] = base64.encodestring(str_comment)
        response = self.bigip.icr_session.patch(
            request_url, data=json.dumps(payload),
            timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return True
        else:
            Log.error('device', response.text)
            raise exceptions.DeviceUpdateException(response.text)
        return False

    @log
    def get_metadata(self, name=None):
        """Get device metadata """
        if not name:
            name = self.get_device_name()
        request_url = self.bigip.icr_url + '/cm/device/~Common~'
        request_url += name + '?$select=name,description'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        str_comment = None
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            if response_obj['name'] == name:
                if 'description' in response_obj:
                    str_comment = response_obj['description']
        elif response.status_code != 404:
            Log.error('device', response.text)
            raise exceptions.DeviceQueryException(response.text)
        if str_comment:
            try:
                return json.loads(base64.decodestring(str_comment))
            except Exception:
                try:
                    return base64.decodestring(str_comment)
                except Exception:
                    return str_comment
        return None

    @log
    def remove_metadata(self, name=None, remove_dict=None):
        """Remove device metadata """
        if not name:
            name = self.get_device_name()
        if isinstance(remove_dict, dict):
            existing_metadata = self.get_metadata(name)
            if isinstance(existing_metadata, dict):
                for key in remove_dict:
                    if key in existing_metadata:
                        del(existing_metadata[key])
                return self.set_metadata(name, existing_metadata)
            else:
                return self.set_metadata(name, '')
        else:
            return self.set_metadata(name, '')

    @log
    def update_metadata(self, name=None, cluster_dict=None):
        """Update device metadata """
        if not name:
            name = self.get_device_name()
        if isinstance(cluster_dict, dict):
            existing_metadata = self.get_metadata(name)
            if isinstance(existing_metadata, dict):
                for key in cluster_dict:
                    existing_metadata[key] = cluster_dict[key]
                return self.set_metadata(name, existing_metadata)
            else:
                return self.set_metadata(name, cluster_dict)
        else:
            return self.set_metadata(name, cluster_dict)



# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip import exceptions
from f5.common import constants as const
from f5.common.logger import Log

import json


class Interface(object):
    """Class for managing iRules on bigip """
    def __init__(self, bigip):
        self.bigip = bigip

    def get_interfaces(self):
        """Get interface names """
        request_url = self.bigip.icr_url + '/net/interface/'
        request_url += '?$select=name'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            names = []
            if 'items' in response_obj:
                for interface in response_obj['items']:
                    names.append(interface['name'])
            return names
        elif response.status_code != 404:
            Log.error('interface', response.text)
            raise exceptions.InterfaceQueryException(response.text)
        return None

    def get_mac_addresses(self):
        """Get MAC addresses for all interfaces """
        request_url = self.bigip.icr_url + '/net/interface/'
        request_url += '?$select=macAddress'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            response_obj = json.loads(response.text)
            macs = []
            if 'items' in response_obj:
                for interface in response_obj['items']:
                    macs.append(interface['macAddress'])
            return macs
        elif response.status_code != 404:
            Log.error('interface', response.text)
            raise exceptions.InterfaceQueryException(response.text)
        return None

    def get_interface_macaddresses_dict(self):
        """Get dictionary of mac addresses keyed by their interface name """
        request_url = self.bigip.icr_url + '/net/interface/'
        request_url += '?$select=name,macAddress'
        response = self.bigip.icr_session.get(
            request_url, timeout=const.CONNECTION_TIMEOUT)
        if response.status_code < 400:
            return_dict = {}
            response_obj = json.loads(response.text)
            if 'items' in response_obj:
                for interface in response_obj['items']:
                    return_dict[interface['name']] = interface['macAddress']
            return return_dict
        elif response.status_code != 404:
            Log.error('interface', response.text)
            raise exceptions.InterfaceQueryException(response.text)
        return None
